import logging
from typing import Dict, Generator, List

from netsquid.components.component import Component, Port
from netsquid.protocols import Protocol

from pydynaa import EventExpression
from qoala.sim.logging import LogManager
from qoala.sim.message import Message


class PortListener(Protocol):
    def __init__(self, port: Port, signal_label: str) -> None:
        self._buffer: List[Message] = []
        self._port: Port = port
        self._signal_label = signal_label
        self.add_signal(signal_label)

    @property
    def buffer(self) -> List[Message]:
        return self._buffer

    def run(self) -> Generator[EventExpression, None, None]:
        while True:
            # Wait for an event saying that there is new input.
            yield self.await_port_input(self._port)

            counter = 0
            # Read all inputs and count them.
            while True:
                input = self._port.rx_input()
                if input is None:
                    break
                self._buffer += input.items
                counter += 1
            # If there are n inputs, there have been n events, but we yielded only
            # on one of them so far. "Flush" these n-1 additional events:
            while counter > 1:
                yield self.await_port_input(self._port)
                counter -= 1

            # Only after having yielded on all current events, we can schedule a
            # notification event, so that its reactor can handle all inputs at once.
            self.send_signal(self._signal_label)


class ComponentProtocol(Protocol):
    def __init__(self, name: str, comp: Component) -> None:
        super().__init__(name)
        self._listeners: Dict[str, PortListener] = {}
        self._logger: logging.Logger = LogManager.get_stack_logger(  # type: ignore
            f"{self.__class__.__name__}({comp.name})"
        )

    def add_listener(self, name, listener: PortListener) -> None:
        self._listeners[name] = listener

    def _receive_msg(
        self, listener_name: str, wake_up_signal: str
    ) -> Generator[EventExpression, None, Message]:
        listener = self._listeners[listener_name]
        if len(listener.buffer) == 0:
            yield self.await_signal(sender=listener, signal_label=wake_up_signal)
        return listener.buffer.pop(0)

    def start(self) -> None:
        super().start()
        for listener in self._listeners.values():
            listener.start()

    def stop(self) -> None:
        for listener in self._listeners.values():
            listener.stop()
        super().stop()



from __future__ import annotations

from netsquid.nodes import Node

from qoala.sim.qnoscomp import QnosComponent


def create_qnoscomp() -> QnosComponent:
    node = Node(name="alice", ID=0)
    return QnosComponent(node)


def test_ports():
    comp = create_qnoscomp()

    assert len(comp.ports) == 6
    assert "host_in" in comp.ports
    assert "host_out" in comp.ports
    assert "nstk_in" in comp.ports
    assert "nstk_out" in comp.ports
    assert "nstk_mem_in" in comp.ports
    assert "nstk_mem_out" in comp.ports


if __name__ == "__main__":
    test_ports()



from __future__ import annotations

from typing import Generator

from pydynaa import EventExpression
from qoala.runtime.environment import LocalEnvironment
from qoala.sim.common import ComponentProtocol, PortListener
from qoala.sim.events import EVENT_WAIT
from qoala.sim.hostcomp import HostComponent
from qoala.sim.message import Message
from qoala.sim.signals import SIGNAL_HAND_HOST_MSG, SIGNAL_HOST_HOST_MSG


class HostInterface(ComponentProtocol):
    """NetSquid protocol representing a Host."""

    def __init__(
        self,
        comp: HostComponent,
        local_env: LocalEnvironment,
    ) -> None:
        """Host protocol constructor.

        :param comp: NetSquid component representing the Host
        """
        name = f"{comp.name}_protocol"
        super().__init__(name=name, comp=comp)
        self._comp = comp

        self._local_env = local_env

        self.add_listener(
            "qnos",
            PortListener(self._comp.ports["qnos_in"], SIGNAL_HAND_HOST_MSG),
        )
        for peer in self._local_env.get_all_other_node_names():
            self.add_listener(
                f"peer_{peer}",
                PortListener(
                    self._comp.peer_in_port(peer), f"{SIGNAL_HOST_HOST_MSG}_{peer}"
                ),
            )

    def send_qnos_msg(self, msg: Message) -> None:
        self._comp.qnos_out_port.tx_output(msg)

    def receive_qnos_msg(self) -> Generator[EventExpression, None, Message]:
        return (yield from self._receive_msg("qnos", SIGNAL_HAND_HOST_MSG))

    def send_peer_msg(self, peer: str, msg: Message) -> None:
        self._comp.peer_out_port(peer).tx_output(msg)

    def receive_peer_msg(self, peer: str) -> Generator[EventExpression, None, Message]:
        return (
            yield from self._receive_msg(
                f"peer_{peer}", f"{SIGNAL_HOST_HOST_MSG}_{peer}"
            )
        )

    def wait(self, delta_time: int) -> Generator[EventExpression, None, None]:
        self._schedule_after(delta_time, EVENT_WAIT)
        event_expr = EventExpression(source=self, event_type=EVENT_WAIT)
        yield event_expr



from __future__ import annotations

import logging
from typing import Generator

from netqasm.lang.operand import Register
from netqasm.lang.parsing.text import NetQASMSyntaxError, parse_register

from pydynaa import EventExpression
from qoala.lang import iqoala
from qoala.sim.hostinterface import HostInterface
from qoala.sim.logging import LogManager
from qoala.sim.message import Message
from qoala.sim.process import IqoalaProcess


class HostProcessor:
    """Does not have state itself. Acts on and changes process objects."""

    def __init__(self, interface: HostInterface, asynchronous: bool = False) -> None:
        self._interface = interface
        self._asynchronous = asynchronous

        # TODO: name
        self._name = f"{interface.name}_HostProcessor"
        self._logger: logging.Logger = LogManager.get_stack_logger(  # type: ignore
            f"{self.__class__.__name__}({self._name})"
        )

        # TODO: add way to set instr latency through constructor
        self._instr_latency: float = 0.0
        self._receive_latency: float = 0.0

    def initialize(self, process: IqoalaProcess) -> None:
        host_mem = process.prog_memory.host_mem
        inputs = process.prog_instance.inputs
        for name, value in inputs.values.items():
            host_mem.write(name, value)

    def assign(
        self, process: IqoalaProcess, instr_idx: int
    ) -> Generator[EventExpression, None, None]:
        csockets = process.csockets
        host_mem = process.prog_memory.host_mem
        pid = process.prog_instance.pid
        program = process.prog_instance.program

        instr = program.instructions[instr_idx]

        self._logger.info(f"Interpreting LHR instruction {instr}")
        if isinstance(instr, iqoala.AssignCValueOp):
            value = instr.attributes[0]
            loc = instr.results[0]
            self._logger.info(f"writing {value} to {loc}")
            host_mem.write(loc, value)
        elif isinstance(instr, iqoala.SendCMsgOp):
            csck_id = host_mem.read(instr.arguments[0])
            csck = csockets[csck_id]
            value = host_mem.read(instr.arguments[1])
            self._logger.info(f"sending msg {value}")
            csck.send_int(value)
        elif isinstance(instr, iqoala.ReceiveCMsgOp):
            csck_id = host_mem.read(instr.arguments[0])
            csck = csockets[csck_id]
            msg = yield from csck.recv_int()
            yield from self._interface.wait(self._receive_latency)
            host_mem.write(instr.results[0], msg)
            self._logger.info(f"received msg {msg}")
        elif isinstance(instr, iqoala.AddCValueOp):
            arg0 = host_mem.read(instr.arguments[0])
            arg1 = host_mem.read(instr.arguments[1])
            loc = instr.results[0]
            result = arg0 + arg1
            self._logger.info(f"computing {loc} = {arg0} + {arg1} = {result}")
            host_mem.write(loc, result)
        elif isinstance(instr, iqoala.MultiplyConstantCValueOp):
            arg0 = host_mem.read(instr.arguments[0])
            const = instr.attributes[0]
            assert isinstance(const, int)
            loc = instr.results[0]
            result = arg0 * const
            self._logger.info(f"computing {loc} = {arg0} * {const} = {result}")
            host_mem.write(loc, result)
        elif isinstance(instr, iqoala.BitConditionalMultiplyConstantCValueOp):
            arg0 = host_mem.read(instr.arguments[0])
            cond = host_mem.read(instr.arguments[1])
            const = instr.attributes[0]
            assert isinstance(const, int)
            loc = instr.results[0]
            if cond == 1:
                result = arg0 * const
            else:
                result = arg0
            self._logger.info(f"computing {loc} = {arg0} * {const}^{cond} = {result}")
            host_mem.write(loc, result)
        elif isinstance(instr, iqoala.RunSubroutineOp):
            arg_vec: iqoala.IqoalaVector = instr.arguments[0]
            args = arg_vec.values
            subrt_name = instr.attributes[0]
            assert isinstance(subrt_name, str)

            iqoala_subrt = process.subroutines[subrt_name]
            self._logger.info(f"executing subroutine {iqoala_subrt}")

            arg_values = {arg: host_mem.read(arg) for arg in args}

            self._logger.info(f"instantiating subroutine with values {arg_values}")
            iqoala_subrt.subroutine.instantiate(pid, arg_values)

            if self._asynchronous:
                # Send a message to Qnos asking it to execute the subroutine.
                self._interface.send_qnos_msg(Message(subrt_name))
                # Wait for Qnos to finish.
                yield from self._interface.receive_qnos_msg()
                # Qnos should have updated the shared memory with subroutine results.
                self.copy_subroutine_results(process, subrt_name)
            else:
                # Let the scheduler make sure that Qnos executes the subroutine at
                # some point. The scheduler is also responsible for copying subroutine
                # results to the Host memory.
                pass

        elif isinstance(instr, iqoala.ReturnResultOp):
            loc = instr.arguments[0]
            value = host_mem.read(loc)
            self._logger.info(f"returning {loc} = {value}")
            process.result.values[loc] = value

        yield from self._interface.wait(self._instr_latency)

    def copy_subroutine_results(self, process: IqoalaProcess, subrt_name: str) -> None:
        iqoala_subrt = process.subroutines[subrt_name]

        for key, mem_loc in iqoala_subrt.return_map.items():
            try:
                reg: Register = parse_register(mem_loc.loc)
                value = process.shared_mem.get_register(reg)
                self._logger.debug(
                    f"writing shared memory value {value} from location "
                    f"{mem_loc} to variable {key}"
                )
                # print(f"subrt result {key} = {value}")
                process.host_mem.write(key, value)
            except NetQASMSyntaxError:
                pass  # TODO: needed?

    @property
    def instr_latency(self) -> float:
        return self._instr_latency

    @instr_latency.setter
    def instr_latency(self, latency: float) -> None:
        self._instr_latency = latency

    @property
    def receive_latency(self) -> float:
        return self._receive_latency

    @receive_latency.setter
    def receive_latency(self, latency: float) -> None:
        self._receive_latency = latency



from __future__ import annotations

from typing import Dict, Optional

from netsquid.components import QuantumProcessor
from netsquid.components.component import Port
from netsquid.nodes import Node

from qoala.runtime.environment import GlobalEnvironment
from qoala.sim.hostcomp import HostComponent
from qoala.sim.netstack import NetstackComponent
from qoala.sim.qnoscomp import QnosComponent


class ProcNodeComponent(Node):
    """NetSquid component representing a quantum network node containing a software
    stack consisting of Host, QNodeOS and QDevice.

    This component has three subcomponents:
        - a QnosComponent
        - a HostComponent

    Has communications ports between
     - the Host component on this node and the Host components on other nodes
     - the QNodeOS component on this node and the QNodeOS components on other nodes

    This is a static container for components and ports.
    Behavior of the node is modeled in the `ProcNode` class, which is a subclass
    of `Protocol`.

    This class is a subclass of the NetSquid `Node` class and can hence be used as
    a standard NetSquid node.
    """

    def __init__(
        self,
        name: str,
        qprocessor: QuantumProcessor,
        global_env: GlobalEnvironment,
        node_id: Optional[int] = None,
    ) -> None:
        """ProcNodeComponent constructor. Typically created indirectly through
        constructing a `ProcNode`."""
        super().__init__(name, ID=node_id)
        self.qmemory = qprocessor

        qnos_comp = QnosComponent(self)
        self.add_subcomponent(qnos_comp, "qnos")

        host_comp = HostComponent(self, global_env)
        self.add_subcomponent(host_comp, "host")

        netstack_comp = NetstackComponent(self, global_env)
        self.add_subcomponent(netstack_comp, "netstack")

        self.host_comp.ports["qnos_out"].connect(self.qnos_comp.ports["host_in"])
        self.host_comp.ports["qnos_in"].connect(self.qnos_comp.ports["host_out"])

        # Ports for communicating with other nodes
        self._netstack_peer_in_ports: Dict[str, str] = {}  # peer name -> port name
        self._netstack_peer_out_ports: Dict[str, str] = {}  # peer name -> port name
        self._host_peer_in_ports: Dict[str, str] = {}  # peer name -> port name
        self._host_peer_out_ports: Dict[str, str] = {}  # peer name -> port name

        for other_node in global_env.get_nodes().values():
            if other_node.name == self.name:
                continue

            netstack_port_in_name = f"netstack_peer_{other_node.name}_in"
            netstack_port_out_name = f"netstack_peer_{other_node.name}_out"
            self._netstack_peer_in_ports[other_node.name] = netstack_port_in_name
            self._netstack_peer_out_ports[other_node.name] = netstack_port_out_name

            host_port_in_name = f"host_peer_{other_node.name}_in"
            host_port_out_name = f"host_peer_{other_node.name}_out"
            self._host_peer_in_ports[other_node.name] = host_port_in_name
            self._host_peer_out_ports[other_node.name] = host_port_out_name

        self.add_ports(self._netstack_peer_in_ports.values())
        self.add_ports(self._netstack_peer_out_ports.values())
        self.add_ports(self._host_peer_in_ports.values())
        self.add_ports(self._host_peer_out_ports.values())

        for other_node in global_env.get_nodes().values():
            if other_node.name == self.name:
                continue
            self.netstack_comp.peer_out_port(other_node.name).forward_output(
                self.netstack_peer_out_port(other_node.name)
            )
            self.netstack_peer_in_port(other_node.name).forward_input(
                self.netstack_comp.peer_in_port(other_node.name)
            )
            self.host_comp.peer_out_port(other_node.name).forward_output(
                self.host_peer_out_port(other_node.name)
            )
            self.host_peer_in_port(other_node.name).forward_input(
                self.host_comp.peer_in_port(other_node.name)
            )

    @property
    def host_comp(self) -> HostComponent:
        return self.subcomponents["host"]

    @property
    def qnos_comp(self) -> QnosComponent:
        return self.subcomponents["qnos"]

    @property
    def netstack_comp(self) -> NetstackComponent:
        return self.subcomponents["netstack"]

    @property
    def qprocessor(self) -> QuantumProcessor:
        return self.qmemory

    def host_peer_in_port(self, name: str) -> Port:
        port_name = self._host_peer_in_ports[name]
        return self.ports[port_name]

    def host_peer_out_port(self, name: str) -> Port:
        port_name = self._host_peer_out_ports[name]
        return self.ports[port_name]

    def netstack_peer_in_port(self, name: str) -> Port:
        port_name = self._netstack_peer_in_ports[name]
        return self.ports[port_name]

    def netstack_peer_out_port(self, name: str) -> Port:
        port_name = self._netstack_peer_out_ports[name]
        return self.ports[port_name]



from __future__ import annotations

from typing import Any, Dict, List

import netsquid as ns

from qoala.runtime.config import ProcNodeNetworkConfig
from qoala.runtime.context import SimulationContext
from qoala.runtime.environment import GlobalEnvironment
from qoala.runtime.program import ProgramInstance
from qoala.sim.build import build_network
from qoala.sim.globals import GlobalSimData
from qoala.sim.network import ProcNodeNetwork


def _run(network: ProcNodeNetwork) -> List[Dict[str, Any]]:
    """Run the protocols of a network and programs running in that network.

    :param network: `ProcNodeNetwork` representing the nodes and links
    :return: final results of the programs
    """

    # Start the link protocols.
    for link in network.links:
        link.start()

    # Start the node protocols.
    for _, node in network.nodes.items():
        node.start()

    # Start the NetSquid simulation.
    ns.sim_run()

    return [node.scheduler.get_results() for _, node in network.nodes.items()]


def run(
    config: ProcNodeNetworkConfig,
    programs: Dict[str, List[ProgramInstance]],
    # schedules: Optional[Dict[str, Schedule]] = None,
    num_times: int = 1,
) -> List[Dict[str, Any]]:
    """Run programs on a network specified by a network configuration.

    :param config: configuration of the network
    :param programs: dictionary of node names to programs
    :param num_times: numbers of times to run the programs, defaults to 1
    :return: program results
    """
    # Create global runtime environment.
    rte = GlobalEnvironment()

    # Build the network. Info about created nodes will be added to the runtime environment.
    network = build_network(config, rte)

    ###########################################
    # TODO: pass context to simulation objects!
    ###########################################
    sim_data = GlobalSimData()
    sim_data.set_network(network)
    context = SimulationContext(global_env=rte, global_sim_data=sim_data)
    print(context)

    for name, program_list in programs.items():
        for program in program_list:
            network.nodes[name]._local_env.register_program(program)

    # if schedules is not None:
    #     for name, schedule in schedules.items():
    #         network.nodes[name]._local_env.install_local_schedule(schedule)

    for name in programs.keys():
        network.nodes[name].install_environment()

    results = _run(network)
    return results



from dataclasses import dataclass


@dataclass
class EprSocket:
    """EPR Socket. Allows for EPR pair generation with a single remote node.

    Multiple EPR Sockets may be created for a single pair of nodes. These
    sockets have a different ID, and may e.g be used for EPR generation requests
    with different parameters."""

    socket_id: int
    remote_id: int
    fidelity: float



from typing import Dict, Generator, List, Optional

import pytest
from netqasm.sdk.build_epr import (
    SER_RESPONSE_KEEP_IDX_BELL_STATE,
    SER_RESPONSE_KEEP_IDX_GOODNESS,
    SER_RESPONSE_KEEP_LEN,
)
from netsquid.components.instructions import INSTR_ROT_X
from netsquid.qubits.ketstates import BellIndex
from qlink_interface import (
    ReqCreateAndKeep,
    ReqCreateBase,
    ReqMeasureDirectly,
    ReqReceive,
    ReqRemoteStatePrep,
    ResCreateAndKeep,
)
from qlink_interface.interface import ResCreate

from pydynaa import EventExpression
from qoala.lang.iqoala import IqoalaProgram, ProgramMeta
from qoala.runtime.program import ProgramInput, ProgramInstance, ProgramResult
from qoala.runtime.schedule import ProgramTaskList
from qoala.sim.constants import PI
from qoala.sim.memmgr import AllocError, MemoryManager
from qoala.sim.memory import ProgramMemory, Topology, UnitModule
from qoala.sim.message import Message
from qoala.sim.netstackinterface import NetstackInterface
from qoala.sim.netstackprocessor import NetstackProcessor
from qoala.sim.process import IqoalaProcess
from qoala.sim.qdevice import PhysicalQuantumMemory, QDevice, QDeviceCommand
from qoala.sim.requests import (
    EprCreateType,
    NetstackCreateRequest,
    NetstackReceiveRequest,
)
from qoala.util.tests import netsquid_run


class MockNetstackInterface(NetstackInterface):
    def __init__(
        self,
        qdevice: QDevice,
        memmgr: MemoryManager,
        mock_result: ResCreate,
    ) -> None:
        self._qdevice = qdevice
        self._memmgr = memmgr
        self._mock_result = mock_result

        self._requests_put: Dict[int, List[ReqCreateBase]] = {}
        self._awaited_result_ck: List[int] = []  # list of remote ids
        self._awaited_mem_free_sig_count: int = 0

    def put_request(self, remote_id: int, request: ReqCreateBase) -> None:
        if remote_id not in self._requests_put:
            self._requests_put[remote_id] = []
        self._requests_put[remote_id].append(request)

    def await_result_create_keep(
        self, remote_id: int
    ) -> Generator[EventExpression, None, ResCreateAndKeep]:
        self._awaited_result_ck.append(remote_id)
        return self._mock_result
        yield  # to make this behave as a generator

    def await_memory_freed_signal(
        self, pid: int, virt_id: int
    ) -> Generator[EventExpression, None, None]:
        raise AllocError
        yield  # to make this behave as a generator

    def send_qnos_msg(self, msg: Message) -> None:
        pass

    def reset(self) -> None:
        self._requests_put = {}
        self._awaited_result_ck = []
        self._awaited_mem_free_sig_count = 0


class MockQDevice(QDevice):
    def __init__(self, topology: Topology) -> None:
        self._memory = PhysicalQuantumMemory(topology.comm_ids, topology.mem_ids)

        self._executed_commands: List[QDeviceCommand] = []

    def set_mem_pos_in_use(self, id: int, in_use: bool) -> None:
        pass

    def execute_commands(
        self, commands: List[QDeviceCommand]
    ) -> Generator[EventExpression, None, Optional[int]]:
        self._executed_commands.extend(commands)
        return None
        yield

    def reset(self) -> None:
        self._executed_commands = []


def create_process(pid: int, unit_module: UnitModule) -> IqoalaProcess:
    program = IqoalaProgram(
        instructions=[], subroutines={}, meta=ProgramMeta.empty("prog")
    )
    instance = ProgramInstance(
        pid=pid,
        program=program,
        inputs=ProgramInput({}),
        tasks=ProgramTaskList.empty(program),
    )
    mem = ProgramMemory(pid=pid, unit_module=unit_module)

    process = IqoalaProcess(
        prog_instance=instance,
        prog_memory=mem,
        csockets={},
        epr_sockets=program.meta.epr_sockets,
        subroutines=program.subroutines,
        requests={},
        result=ProgramResult(values={}),
    )
    return process


def test_create_link_layer_create_request():
    qdevice = MockQDevice(Topology(comm_ids={0}, mem_ids={1}))
    memmgr = MemoryManager("alice", qdevice)
    mock_result = ResCreateAndKeep(bell_state=BellIndex.B00)
    interface = MockNetstackInterface(qdevice, memmgr, mock_result)
    processor = NetstackProcessor(interface)

    remote_id = 3
    num_pairs = 2
    fidelity = 0.75
    result_array = 5

    request_ck = NetstackCreateRequest(
        remote_id=remote_id,
        epr_socket_id=0,
        typ=EprCreateType.CREATE_KEEP,
        num_pairs=num_pairs,
        fidelity=fidelity,
        virt_qubit_ids=[0],
        result_array_addr=result_array,
    )

    ll_request_ck = processor._create_link_layer_create_request(request_ck)

    assert ll_request_ck == ReqCreateAndKeep(
        remote_node_id=remote_id, minimum_fidelity=fidelity, number=num_pairs
    )

    request_md = NetstackCreateRequest(
        remote_id=remote_id,
        epr_socket_id=0,
        typ=EprCreateType.MEASURE_DIRECTLY,
        num_pairs=num_pairs,
        fidelity=fidelity,
        virt_qubit_ids=[0],
        result_array_addr=result_array,
    )

    ll_request_md = processor._create_link_layer_create_request(request_md)

    assert ll_request_md == ReqMeasureDirectly(
        remote_node_id=remote_id, minimum_fidelity=fidelity, number=num_pairs
    )

    request_rsp = NetstackCreateRequest(
        remote_id=remote_id,
        epr_socket_id=0,
        typ=EprCreateType.REMOTE_STATE_PREP,
        num_pairs=num_pairs,
        fidelity=fidelity,
        virt_qubit_ids=[0],
        result_array_addr=result_array,
    )

    ll_request_rsp = processor._create_link_layer_create_request(request_rsp)

    assert ll_request_rsp == ReqRemoteStatePrep(
        remote_node_id=remote_id, minimum_fidelity=fidelity, number=num_pairs
    )


def test_create_single_pair_1():
    topology = Topology(comm_ids={0}, mem_ids={1})
    qdevice = MockQDevice(topology)
    unit_module = UnitModule.from_topology(topology)
    memmgr = MemoryManager("alice", qdevice)
    mock_result = ResCreateAndKeep(bell_state=BellIndex.B00)
    interface = MockNetstackInterface(qdevice, memmgr, mock_result)
    processor = NetstackProcessor(interface)

    remote_id = 1
    num_pairs = 3
    fidelity = 0.75
    result_array_addr = 0

    request = NetstackCreateRequest(
        remote_id=remote_id,
        epr_socket_id=0,
        typ=EprCreateType.CREATE_KEEP,
        num_pairs=num_pairs,
        fidelity=fidelity,
        virt_qubit_ids=[1],
        result_array_addr=result_array_addr,
    )

    pid = 0
    process = create_process(pid, unit_module)
    memmgr.add_process(process)

    process.prog_memory.shared_mem.init_new_array(result_array_addr, 10)

    # At the start, nothing should be allocated yet.
    assert memmgr.phys_id_for(pid, virt_id=0) is None

    # Create a single pair (communication with link layer is mocked) for the request.
    netsquid_run(processor.create_single_pair(process, request, virt_id=0))

    # Should have allocated virt_id 0.
    assert memmgr.phys_id_for(pid, virt_id=0) == 0

    assert len(interface._requests_put[remote_id]) == 1

    # Should have put a LL request with correct params **and number of pairs = 1**.
    assert interface._requests_put[remote_id][0] == ReqCreateAndKeep(
        remote_node_id=remote_id, minimum_fidelity=fidelity, number=1
    )

    # Mock result indicated bell state B00, so no local gates should have been executed.
    assert len(qdevice._executed_commands) == 0

    with pytest.raises(AllocError):
        # AllocError since the only comm qubit has already been allocated.
        netsquid_run(processor.create_single_pair(process, request, virt_id=0))

    # Free up the communication qubit again.
    memmgr.free(pid, 0)

    interface.reset()
    qdevice.reset()

    # See the next mocked result to indicate Bell state B01.
    interface._mock_result = ResCreateAndKeep(bell_state=BellIndex.B01)

    # Create a single pair again. Try virt ID 1, which is not a comm qubit.
    with pytest.raises(AllocError):
        netsquid_run(processor.create_single_pair(process, request, virt_id=1))

    # Try with virt ID = 0 again.
    netsquid_run(processor.create_single_pair(process, request, virt_id=0))

    # Should have allocated virt_id 0 again.
    assert memmgr.phys_id_for(pid, virt_id=0) == 0

    assert len(interface._requests_put[remote_id]) == 1

    # Should have put a LL request with correct params **and number of pairs = 1**.
    assert interface._requests_put[remote_id][0] == ReqCreateAndKeep(
        remote_node_id=remote_id, minimum_fidelity=fidelity, number=1
    )

    # Mock result indicated bell state B01, so an X gate should have been executed.
    assert len(qdevice._executed_commands) == 1
    assert qdevice._executed_commands[0] == QDeviceCommand(INSTR_ROT_X, [0], angle=PI)


def test_create_single_pair_2():

    # Let HW have 3 comm qubits.
    topology = Topology(comm_ids={0, 1, 2}, mem_ids={0, 1, 2})
    qdevice = MockQDevice(topology)
    unit_module = UnitModule.from_topology(topology)
    memmgr = MemoryManager("alice", qdevice)
    mock_result = ResCreateAndKeep(bell_state=BellIndex.B00)
    interface = MockNetstackInterface(qdevice, memmgr, mock_result)
    processor = NetstackProcessor(interface)

    remote_id = 1
    num_pairs = 3
    fidelity = 0.75
    result_array_addr = 0

    request = NetstackCreateRequest(
        remote_id=remote_id,
        epr_socket_id=0,
        typ=EprCreateType.CREATE_KEEP,
        num_pairs=num_pairs,
        fidelity=fidelity,
        virt_qubit_ids=[1],
        result_array_addr=result_array_addr,
    )

    process0 = create_process(pid=0, unit_module=unit_module)
    process1 = create_process(pid=1, unit_module=unit_module)
    process2 = create_process(pid=2, unit_module=unit_module)
    memmgr.add_process(process0)
    memmgr.add_process(process1)
    memmgr.add_process(process2)
    process0.prog_memory.shared_mem.init_new_array(result_array_addr, 10)
    process1.prog_memory.shared_mem.init_new_array(result_array_addr, 10)
    process2.prog_memory.shared_mem.init_new_array(result_array_addr, 10)

    # At the start, nothing should be allocated yet.
    assert all(memmgr.phys_id_for(pid=0, virt_id=i) is None for i in range(3))
    assert all(memmgr.phys_id_for(pid=1, virt_id=i) is None for i in range(3))
    assert all(memmgr.phys_id_for(pid=2, virt_id=i) is None for i in range(3))

    # We are use the same request object for all 3 processes.

    # Create a single pair for process 0, based on the request object.
    # Store the result in virt qubit 0.
    netsquid_run(processor.create_single_pair(process0, request, virt_id=0))

    # Should have allocated virt_id 0 for process 0.
    assert memmgr.phys_id_for(pid=0, virt_id=0) == 0
    assert len(interface._requests_put[remote_id]) == 1

    # Create another pair for process 0.
    with pytest.raises(AllocError):
        # Using virt ID 0 again should raise an AllocError.
        netsquid_run(processor.create_single_pair(process0, request, virt_id=0))
    # Use virt ID 1.
    netsquid_run(processor.create_single_pair(process0, request, virt_id=1))

    # Should have allocated virt_id 1 for process 0.
    assert memmgr.phys_id_for(pid=0, virt_id=1) == 1

    # Create a single pair for process 1, using virt ID 0.
    # Should get mapped to phys ID 2 (since 0 and 1 already in use).
    netsquid_run(processor.create_single_pair(process1, request, virt_id=0))
    assert memmgr.phys_id_for(pid=1, virt_id=0) == 2

    # Create a single pair for process 1, using virt ID 1.
    # Should give an AllocError since no physical comm qubits are available.
    with pytest.raises(AllocError):
        netsquid_run(processor.create_single_pair(process1, request, virt_id=1))

    # Create a single pair for process 2, using virt ID 0.
    # Should give an AllocError since no physical comm qubits are available.
    with pytest.raises(AllocError):
        netsquid_run(processor.create_single_pair(process2, request, virt_id=0))


def test_write_pair_result():
    topology = Topology(comm_ids={0, 1, 2}, mem_ids={0, 1, 2})
    qdevice = MockQDevice(topology)
    unit_module = UnitModule.from_topology(topology)
    memmgr = MemoryManager("alice", qdevice)
    mock_result = ResCreateAndKeep(bell_state=BellIndex.B00)
    interface = MockNetstackInterface(qdevice, memmgr, mock_result)
    processor = NetstackProcessor(interface)

    process = create_process(0, unit_module)
    memmgr.add_process(process)

    # Only goodness and bell state should be written to the NetQASM array.
    goodness = 27
    bell_state = 3
    ll_result = ResCreateAndKeep(goodness=goodness, bell_state=bell_state)
    pair_index = 5
    result_array_addr = 0
    duration = 7600.0

    # Make sure the result array is initialized with enough entries.
    # Need at least SER_RESPONSE_KEEP_LEN * num_pairs entries.
    shared_mem = process.prog_memory.shared_mem
    shared_mem.init_new_array(result_array_addr, SER_RESPONSE_KEEP_LEN * 10)

    processor.write_pair_result(
        process, ll_result, pair_index, result_array_addr, duration
    )

    # 'goodness' entry is used for rounded duration in us
    expected_duration = int(duration / 1000)

    assert (
        shared_mem.get_array_part(
            result_array_addr,
            SER_RESPONSE_KEEP_LEN * pair_index + SER_RESPONSE_KEEP_IDX_GOODNESS,
        )
        == expected_duration
    )

    assert (
        shared_mem.get_array_part(
            result_array_addr,
            SER_RESPONSE_KEEP_LEN * pair_index + SER_RESPONSE_KEEP_IDX_BELL_STATE,
        )
        == bell_state
    )


def test_handle_create_ck_request():
    topology = Topology(comm_ids={0, 1, 2}, mem_ids={0, 1, 2})
    qdevice = MockQDevice(topology)
    unit_module = UnitModule.from_topology(topology)
    memmgr = MemoryManager("alice", qdevice)
    mock_result = ResCreateAndKeep(bell_state=BellIndex.B00)
    interface = MockNetstackInterface(qdevice, memmgr, mock_result)
    processor = NetstackProcessor(interface)

    remote_id = 1
    num_pairs = 3
    fidelity = 0.75
    result_array_addr = 0

    request = NetstackCreateRequest(
        remote_id=remote_id,
        epr_socket_id=0,
        typ=EprCreateType.CREATE_KEEP,
        num_pairs=num_pairs,
        fidelity=fidelity,
        virt_qubit_ids=[0, 1, 2],
        result_array_addr=result_array_addr,
    )

    process = create_process(0, unit_module)
    memmgr.add_process(process)
    shared_mem = process.prog_memory.shared_mem
    shared_mem.init_new_array(result_array_addr, SER_RESPONSE_KEEP_LEN * num_pairs)

    netsquid_run(processor.handle_create_ck_request(process, request))
    assert memmgr.phys_id_for(pid=0, virt_id=0) == 0
    assert memmgr.phys_id_for(pid=0, virt_id=1) == 1
    assert memmgr.phys_id_for(pid=0, virt_id=2) == 2


def test_handle_create_ck_request_invalid_virt_ids():
    topology = Topology(comm_ids={0, 1, 2}, mem_ids={0, 1, 2})
    qdevice = MockQDevice(topology)
    unit_module = UnitModule.from_topology(topology)
    memmgr = MemoryManager("alice", qdevice)
    mock_result = ResCreateAndKeep(bell_state=BellIndex.B00)
    interface = MockNetstackInterface(qdevice, memmgr, mock_result)
    processor = NetstackProcessor(interface)

    remote_id = 1
    num_pairs = 3
    fidelity = 0.75
    result_array_addr = 0

    request = NetstackCreateRequest(
        remote_id=remote_id,
        epr_socket_id=0,
        typ=EprCreateType.CREATE_KEEP,
        num_pairs=num_pairs,
        fidelity=fidelity,
        virt_qubit_ids=[0, 0, 0],
        result_array_addr=result_array_addr,
    )

    process = create_process(0, unit_module)
    memmgr.add_process(process)
    shared_mem = process.prog_memory.shared_mem
    shared_mem.init_new_array(result_array_addr, SER_RESPONSE_KEEP_LEN * num_pairs)

    with pytest.raises(AllocError):
        # After the first pair has been generated, virt ID 0 has been allocated,
        # so the next pair should not be able to allocate virt ID 0 again.
        netsquid_run(processor.handle_create_ck_request(process, request))


def test_receive_single_pair_1():
    topology = Topology(comm_ids={0}, mem_ids={1})
    qdevice = MockQDevice(topology)
    unit_module = UnitModule.from_topology(topology)
    memmgr = MemoryManager("alice", qdevice)
    mock_result = ResCreateAndKeep(bell_state=BellIndex.B00)
    interface = MockNetstackInterface(qdevice, memmgr, mock_result)
    processor = NetstackProcessor(interface)

    remote_id = 1
    num_pairs = 3
    fidelity = 0.75
    result_array_addr = 0

    request = NetstackReceiveRequest(
        remote_id=remote_id,
        epr_socket_id=0,
        typ=EprCreateType.CREATE_KEEP,
        num_pairs=num_pairs,
        fidelity=fidelity,
        virt_qubit_ids=[1],
        result_array_addr=result_array_addr,
    )

    pid = 0
    process = create_process(pid, unit_module)
    memmgr.add_process(process)

    process.prog_memory.shared_mem.init_new_array(result_array_addr, 10)

    # At the start, nothing should be allocated yet.
    assert memmgr.phys_id_for(pid, virt_id=0) is None

    # Create a single pair (communication with link layer is mocked) for the request.
    netsquid_run(processor.receive_single_pair(process, request, virt_id=0))

    # Should have allocated virt_id 0.
    assert memmgr.phys_id_for(pid, virt_id=0) == 0

    assert len(interface._requests_put[remote_id]) == 1

    # Should have put a LL request with correct params **and number of pairs = 1**.
    assert interface._requests_put[remote_id][0] == ReqReceive(remote_node_id=remote_id)

    # No local gates should have been executed.
    assert len(qdevice._executed_commands) == 0

    with pytest.raises(AllocError):
        # AllocError since the only comm qubit has already been allocated.
        netsquid_run(processor.receive_single_pair(process, request, virt_id=0))

    # Free up the communication qubit again.
    memmgr.free(pid, 0)

    interface.reset()
    qdevice.reset()

    # See the next mocked result to indicate Bell state B01.
    interface._mock_result = ResCreateAndKeep(bell_state=BellIndex.B01)

    # Create a single pair again. Try virt ID 1, which is not a comm qubit.
    with pytest.raises(AllocError):
        netsquid_run(processor.receive_single_pair(process, request, virt_id=1))

    # Try with virt ID = 0 again.
    netsquid_run(processor.receive_single_pair(process, request, virt_id=0))

    # Should have allocated virt_id 0 again.
    assert memmgr.phys_id_for(pid, virt_id=0) == 0

    assert len(interface._requests_put[remote_id]) == 1

    # Should have put a LL request with correct params **and number of pairs = 1**.
    assert interface._requests_put[remote_id][0] == ReqReceive(remote_node_id=remote_id)

    # Still no local gates should have been executed.
    assert len(qdevice._executed_commands) == 0


def test_receive_single_pair_2():

    # Let HW have 3 comm qubits.
    topology = Topology(comm_ids={0, 1, 2}, mem_ids={0, 1, 2})
    qdevice = MockQDevice(topology)
    unit_module = UnitModule.from_topology(topology)
    memmgr = MemoryManager("alice", qdevice)
    mock_result = ResCreateAndKeep(bell_state=BellIndex.B00)
    interface = MockNetstackInterface(qdevice, memmgr, mock_result)
    processor = NetstackProcessor(interface)

    remote_id = 1
    num_pairs = 3
    fidelity = 0.75
    result_array_addr = 0

    request = NetstackReceiveRequest(
        remote_id=remote_id,
        epr_socket_id=0,
        typ=EprCreateType.CREATE_KEEP,
        num_pairs=num_pairs,
        fidelity=fidelity,
        virt_qubit_ids=[1],
        result_array_addr=result_array_addr,
    )

    process0 = create_process(pid=0, unit_module=unit_module)
    process1 = create_process(pid=1, unit_module=unit_module)
    process2 = create_process(pid=2, unit_module=unit_module)
    memmgr.add_process(process0)
    memmgr.add_process(process1)
    memmgr.add_process(process2)
    process0.prog_memory.shared_mem.init_new_array(result_array_addr, 10)
    process1.prog_memory.shared_mem.init_new_array(result_array_addr, 10)
    process2.prog_memory.shared_mem.init_new_array(result_array_addr, 10)

    # At the start, nothing should be allocated yet.
    assert all(memmgr.phys_id_for(pid=0, virt_id=i) is None for i in range(3))
    assert all(memmgr.phys_id_for(pid=1, virt_id=i) is None for i in range(3))
    assert all(memmgr.phys_id_for(pid=2, virt_id=i) is None for i in range(3))

    # We are use the same request object for all 3 processes.

    # Create a single pair for process 0, based on the request object.
    # Store the result in virt qubit 0.
    netsquid_run(processor.receive_single_pair(process0, request, virt_id=0))

    # Should have allocated virt_id 0 for process 0.
    assert memmgr.phys_id_for(pid=0, virt_id=0) == 0
    assert len(interface._requests_put[remote_id]) == 1

    # Create another pair for process 0.
    with pytest.raises(AllocError):
        # Using virt ID 0 again should raise an AllocError.
        netsquid_run(processor.receive_single_pair(process0, request, virt_id=0))
    # Use virt ID 1.
    netsquid_run(processor.receive_single_pair(process0, request, virt_id=1))

    # Should have allocated virt_id 1 for process 0.
    assert memmgr.phys_id_for(pid=0, virt_id=1) == 1

    # Create a single pair for process 1, using virt ID 0.
    # Should get mapped to phys ID 2 (since 0 and 1 already in use).
    netsquid_run(processor.receive_single_pair(process1, request, virt_id=0))
    assert memmgr.phys_id_for(pid=1, virt_id=0) == 2

    # Create a single pair for process 1, using virt ID 1.
    # Should give an AllocError since no physical comm qubits are available.
    with pytest.raises(AllocError):
        netsquid_run(processor.receive_single_pair(process1, request, virt_id=1))

    # Create a single pair for process 2, using virt ID 0.
    # Should give an AllocError since no physical comm qubits are available.
    with pytest.raises(AllocError):
        netsquid_run(processor.receive_single_pair(process2, request, virt_id=0))


def test_handle_receive_ck_request():
    topology = Topology(comm_ids={0, 1, 2}, mem_ids={0, 1, 2})
    qdevice = MockQDevice(topology)
    unit_module = UnitModule.from_topology(topology)
    memmgr = MemoryManager("alice", qdevice)
    mock_result = ResCreateAndKeep(bell_state=BellIndex.B00)
    interface = MockNetstackInterface(qdevice, memmgr, mock_result)
    processor = NetstackProcessor(interface)

    remote_id = 1
    num_pairs = 3
    fidelity = 0.75
    result_array_addr = 0

    request = NetstackReceiveRequest(
        remote_id=remote_id,
        epr_socket_id=0,
        typ=EprCreateType.CREATE_KEEP,
        num_pairs=num_pairs,
        fidelity=fidelity,
        virt_qubit_ids=[0, 1, 2],
        result_array_addr=result_array_addr,
    )

    process = create_process(0, unit_module)
    memmgr.add_process(process)
    shared_mem = process.prog_memory.shared_mem
    shared_mem.init_new_array(result_array_addr, SER_RESPONSE_KEEP_LEN * num_pairs)

    netsquid_run(processor.handle_receive_ck_request(process, request))
    assert memmgr.phys_id_for(pid=0, virt_id=0) == 0
    assert memmgr.phys_id_for(pid=0, virt_id=1) == 1
    assert memmgr.phys_id_for(pid=0, virt_id=2) == 2


if __name__ == "__main__":
    test_create_link_layer_create_request()
    test_create_single_pair_1()
    test_create_single_pair_2()
    test_write_pair_result()
    test_handle_create_ck_request()
    test_handle_create_ck_request_invalid_virt_ids()
    test_receive_single_pair_1()
    test_receive_single_pair_2()
    test_handle_receive_ck_request()



# from qoala.lang.iqoala import IqoalaParser
# from qoala.runtime.config import GenericQDeviceConfig, LinkConfig
# from qoala.runtime.program import ProgramInstance
# from qoala.runtime.run import run
# from sim.stack.common import LogManager
# from sim.stack.program import ProgramMeta

# def test_run():
#     program_text = """
# new_value = assign_cval() : 8
# my_value = add_cval_c(new_value, new_value)
# run_subroutine(vec<my_value>) :
#     return M0 -> m
#   NETQASM_START
#     set Q0 0
#     qalloc Q0
#     init Q0
#     rot_x Q0 {my_value} 4
#     meas Q0 M0
#     ret_reg M0
#   NETQASM_END
# return_result(m)
#     """
#     parsed_program = IqoalaParser(program_text).parse()
#     parsed_program.meta = ProgramMeta(
#         name="client", parameters={}, csockets=[], epr_sockets=[], max_qubits=1
#     )

#     sender_stack = StackConfig(
#         name="client",
#         qdevice_typ="generic",
#         qdevice_cfg=GenericQDeviceConfig.perfect_config(),
#     )

#     prog_instance = ProgramInstance(parsed_program, {}, 1, 0.0)

#     cfg = StackNetworkConfig(stacks=[sender_stack], links=[])
#     result = run(cfg, programs={"client": prog_instance})
#     print(result)


# def test_run_two_nodes_classical():
#     program_text_client = """
# new_value = assign_cval() : 8
# send_cmsg(new_value)
#     """
#     program_client = IqoalaParser(program_text_client).parse()
#     program_client.meta = ProgramMeta(
#         name="client", parameters={}, csockets=["server"], epr_sockets=[], max_qubits=1
#     )
#     client_stack = StackConfig(
#         name="client",
#         qdevice_typ="generic",
#         qdevice_cfg=GenericQDeviceConfig.perfect_config(),
#     )

#     program_text_server = """
# new_value = assign_cval() : 8
# value = recv_cmsg()
# return_result(value)
#     """
#     program_server = IqoalaParser(program_text_server).parse()
#     program_server.meta = ProgramMeta(
#         name="client", parameters={}, csockets=["client"], epr_sockets=[], max_qubits=1
#     )
#     server_stack = StackConfig(
#         name="server",
#         qdevice_typ="generic",
#         qdevice_cfg=GenericQDeviceConfig.perfect_config(),
#     )

#     prog_server_instance = ProgramInstance(program_server, {}, 1, 0.0)
#     prog_client_instance = ProgramInstance(program_client, {}, 1, 0.0)

#     cfg = StackNetworkConfig(stacks=[client_stack, server_stack], links=[])
#     result = run(
#         cfg, programs={"client": prog_client_instance, "server": prog_server_instance}
#     )
#     print(result)


# def test_run_two_nodes_epr():
#     program_text_client = """
# remote_id = assign_cval() : 1
# epr_sck_id = assign_cval() : 0
# run_subroutine(vec<remote_id; epr_sck_id>) :
#     return M0 -> m
#   NETQASM_START
#     set R0 0
#     set R1 1
#     set R2 2
#     set R3 20
#     set R10 10
#     set C0 {remote_id}
#     set C1 {epr_sck_id}
#     array R10 @0
#     array R1 @1
#     array R3 @2
#     store R0 @1[R0]
#     store R0 @2[R0]
#     store R1 @2[R1]
#     create_epr C0 C1 R1 R2 R0
#     wait_all @0[R0:R10]
#     set Q0 0
#     meas Q0 M0
#     qfree Q0
#     ret_reg M0
#   NETQASM_END
# return_result(m)
#     """
#     program_client = IqoalaParser(program_text_client).parse()
#     program_client.meta = ProgramMeta(
#         name="client",
#         parameters={},
#         csockets=["server"],
#         epr_sockets=["server"],
#         max_qubits=1,
#     )
#     client_stack = StackConfig(
#         name="client",
#         qdevice_typ="generic",
#         qdevice_cfg=GenericQDeviceConfig.perfect_config(),
#     )

#     program_text_server = """
# run_subroutine(vec<>) :
#     return M0 -> m
#   NETQASM_START
#     set R0 0
#     set R1 1
#     set R2 2
#     set R10 10
#     array R10 @0
#     array R1 @1
#     store R0 @1[R0]
#     recv_epr R0 R0 R1 R0
#     wait_all @0[R0:R10]
#     set Q0 0
#     meas Q0 M0
#     qfree Q0
#     ret_reg M0
#   NETQASM_END
# return_result(m)
#     """
#     program_server = IqoalaParser(program_text_server).parse()
#     program_server.meta = ProgramMeta(
#         name="client",
#         parameters={},
#         csockets=["client"],
#         epr_sockets=["client"],
#         max_qubits=1,
#     )
#     server_stack = StackConfig(
#         name="server",
#         qdevice_typ="generic",
#         qdevice_cfg=GenericQDeviceConfig.perfect_config(),
#     )
#     link = LinkConfig(
#         stack1="client",
#         stack2="server",
#         typ="perfect",
#     )

#     prog_server_instance = ProgramInstance(program_server, {}, 1, 0.0)
#     prog_client_instance = ProgramInstance(program_client, {}, 1, 0.0)

#     cfg = StackNetworkConfig(stacks=[client_stack, server_stack], links=[link])
#     result = run(
#         cfg, programs={"client": prog_client_instance, "server": prog_server_instance}
#     )
#     print(result)


# if __name__ == "__main__":
#     LogManager.set_log_level("DEBUG")
#     test_run()
#     test_run_two_nodes_classical()
#     test_run_two_nodes_epr()



from typing import Generator, List

import netsquid as ns
import numpy as np
from netsquid.protocols import Protocol
from netsquid.qubits import qubitapi
from netsquid.qubits.qubit import Qubit

from pydynaa import EventExpression


def yield_from(generator: Generator):
    try:
        while True:
            next(generator)
    except StopIteration as e:
        return e.value


class SimpleNetSquidProtocol(Protocol):
    def __init__(self, generator: Generator) -> None:
        super().__init__("test")
        self._gen = generator
        self.result = None

    def run(self) -> Generator[EventExpression, None, None]:
        self.result = yield from self._gen


def netsquid_run(generator: Generator):
    prot = SimpleNetSquidProtocol(generator)
    prot.start()
    ns.sim_run()
    return prot.result


def text_equal(text1, text2) -> bool:
    # allows whitespace differences
    lines1 = [line.strip() for line in text1.split("\n") if len(line) > 0]
    lines2 = [line.strip() for line in text2.split("\n") if len(line) > 0]
    for line1, line2 in zip(lines1, lines2):
        if line1 != line2:
            return False
    return True


def has_state(qubit: Qubit, state: np.ndarray, margin: float = 0.001) -> bool:
    return abs(1.0 - qubitapi.fidelity(qubit, state, squared=True)) < margin


def has_multi_state(
    qubits: List[Qubit], state: np.ndarray, margin: float = 0.001
) -> bool:
    return abs(1.0 - qubitapi.fidelity(qubits, state, squared=True)) < margin


def has_max_mixed_state(qubit: Qubit, margin: float = 0.001) -> bool:
    # TODO: fix fidelity calculation with mixed states
    max_mixed = np.array([[0.5, 0], [0, 0.5]])
    # qubit_state = qubitapi.reduced_dm(qubit)
    return np.isclose(qubitapi.reduced_dm(qubit), max_mixed, margin)



from __future__ import annotations

from typing import Generator

import netsquid as ns
from netsquid.nodes import Node

from pydynaa import EventExpression
from qoala.runtime.environment import (
    GlobalEnvironment,
    GlobalNodeInfo,
    LocalEnvironment,
)
from qoala.sim.message import Message
from qoala.sim.netstackcomp import NetstackComponent
from qoala.sim.netstackinterface import NetstackInterface


class MockNetstackInterface(NetstackInterface):
    def __init__(self, comp: NetstackComponent, local_env: LocalEnvironment) -> None:
        super().__init__(comp, local_env, None, None, None)


def create_netstackcomp(num_other_nodes: int) -> NetstackComponent:
    node = Node(name="alice", ID=0)
    env = GlobalEnvironment()

    node_info = GlobalNodeInfo(node.name, node.ID)
    env.add_node(node.ID, node_info)

    for id in range(1, num_other_nodes + 1):
        node_info = GlobalNodeInfo(f"node_{id}", id)
        env.add_node(id, node_info)

    return NetstackComponent(node, env)


def test_no_other_nodes():
    comp = create_netstackcomp(num_other_nodes=0)

    # should have qnos_in, qnos_out, qnos_mem_in, qnos_mem_out ports
    assert len(comp.ports) == 4
    assert "qnos_in" in comp.ports
    assert "qnos_out" in comp.ports
    assert "qnos_mem_in" in comp.ports
    assert "qnos_mem_out" in comp.ports


def test_one_other_node():
    comp = create_netstackcomp(num_other_nodes=1)

    # should have 4 qnos ports + 2 peer ports
    assert len(comp.ports) == 6
    assert "qnos_in" in comp.ports
    assert "qnos_out" in comp.ports
    assert "qnos_mem_in" in comp.ports
    assert "qnos_mem_out" in comp.ports

    assert "peer_node_1_in" in comp.ports
    assert "peer_node_1_out" in comp.ports

    # Test properties
    assert comp.peer_in_port("node_1") == comp.ports["peer_node_1_in"]
    assert comp.peer_out_port("node_1") == comp.ports["peer_node_1_out"]


def test_many_other_nodes():
    comp = create_netstackcomp(num_other_nodes=5)

    # should have 4 qnos ports + 5 * 2 peer ports
    assert len(comp.ports) == 14
    assert "qnos_in" in comp.ports
    assert "qnos_out" in comp.ports
    assert "qnos_mem_in" in comp.ports
    assert "qnos_mem_out" in comp.ports

    for i in range(1, 6):
        assert f"peer_node_{i}_in" in comp.ports
        assert f"peer_node_{i}_out" in comp.ports
        # Test properties
        assert comp.peer_in_port(f"node_{i}") == comp.ports[f"peer_node_{i}_in"]
        assert comp.peer_out_port(f"node_{i}") == comp.ports[f"peer_node_{i}_out"]


def test_connection():
    alice = Node(name="alice", ID=0)
    bob = Node(name="bob", ID=1)
    env = GlobalEnvironment()

    alice_info = GlobalNodeInfo(alice.name, alice.ID)
    env.add_node(alice.ID, alice_info)
    bob_info = GlobalNodeInfo(bob.name, bob.ID)
    env.add_node(bob.ID, bob_info)

    alice_comp = NetstackComponent(alice, env)
    bob_comp = NetstackComponent(bob, env)

    alice_comp.peer_out_port("bob").connect(bob_comp.peer_in_port("alice"))
    alice_comp.peer_in_port("bob").connect(bob_comp.peer_out_port("alice"))

    class AliceNetstackInterface(MockNetstackInterface):
        def run(self) -> Generator[EventExpression, None, None]:
            self.send_peer_msg("bob", Message("hello"))

    class BobNetstackInterface(MockNetstackInterface):
        def run(self) -> Generator[EventExpression, None, None]:
            msg = yield from self.receive_peer_msg("alice")
            assert msg.content == "hello"

    alice_intf = AliceNetstackInterface(alice_comp, LocalEnvironment(env, alice.ID))
    bob_intf = BobNetstackInterface(bob_comp, LocalEnvironment(env, bob.ID))

    alice_intf.start()
    bob_intf.start()

    ns.sim_run()


def test_three_way_connection():
    alice = Node(name="alice", ID=0)
    bob = Node(name="bob", ID=1)
    charlie = Node(name="charlie", ID=2)
    env = GlobalEnvironment()

    alice_info = GlobalNodeInfo(alice.name, alice.ID)
    env.add_node(alice.ID, alice_info)
    bob_info = GlobalNodeInfo(bob.name, bob.ID)
    env.add_node(bob.ID, bob_info)
    charlie_info = GlobalNodeInfo(charlie.name, charlie.ID)
    env.add_node(charlie.ID, charlie_info)

    alice_comp = NetstackComponent(alice, env)
    bob_comp = NetstackComponent(bob, env)
    charlie_comp = NetstackComponent(charlie, env)

    alice_comp.peer_out_port("bob").connect(bob_comp.peer_in_port("alice"))
    alice_comp.peer_in_port("bob").connect(bob_comp.peer_out_port("alice"))
    alice_comp.peer_out_port("charlie").connect(charlie_comp.peer_in_port("alice"))
    alice_comp.peer_in_port("charlie").connect(charlie_comp.peer_out_port("alice"))
    bob_comp.peer_out_port("charlie").connect(charlie_comp.peer_in_port("bob"))
    bob_comp.peer_in_port("charlie").connect(charlie_comp.peer_out_port("bob"))

    class AliceNetstackInterface(MockNetstackInterface):
        def run(self) -> Generator[EventExpression, None, None]:
            self.send_peer_msg("bob", Message("hello bob"))
            self.send_peer_msg("charlie", Message("hello charlie"))

    class BobNetstackInterface(MockNetstackInterface):
        def run(self) -> Generator[EventExpression, None, None]:
            msg = yield from self.receive_peer_msg("alice")
            assert msg.content == "hello bob"

    class CharlieNetstackInterface(MockNetstackInterface):
        def run(self) -> Generator[EventExpression, None, None]:
            msg = yield from self.receive_peer_msg("alice")
            assert msg.content == "hello charlie"

    alice_intf = AliceNetstackInterface(alice_comp, LocalEnvironment(env, alice.ID))
    bob_intf = BobNetstackInterface(bob_comp, LocalEnvironment(env, bob.ID))
    charlie_intf = CharlieNetstackInterface(
        charlie_comp, LocalEnvironment(env, charlie.ID)
    )

    alice_intf.start()
    bob_intf.start()
    charlie_intf.start()

    ns.sim_run()


if __name__ == "__main__":
    test_no_other_nodes()
    test_one_other_node()
    test_many_other_nodes()

    test_connection()
    test_three_way_connection()



from qoala.sim.memory import SharedMemory


def test_register_setup():
    mem = SharedMemory(0)
    print(mem._registers)

    mem.set_reg_value("R0", 42)
    print(mem.get_reg_value("R0"))

    print(mem._registers)


if __name__ == "__main__":
    test_register_setup()



from __future__ import annotations

from dataclasses import dataclass
from typing import Dict, List, Optional, Set, Tuple, Type, Union

from netqasm.lang import operand
from netqasm.lang.encoding import RegisterName
from netqasm.sdk.shared_memory import Arrays, RegisterGroup, setup_registers


class RegisterMeta:
    @classmethod
    def prefixes(cls) -> List[str]:
        return ["R", "C", "Q", "M"]

    @classmethod
    def parse(cls, name: str) -> Tuple[RegisterName, int]:
        assert len(name) >= 2
        assert name[0] in cls.prefixes()
        group = RegisterName[name[0]]
        index = int(name[1:])
        assert index < 16
        return group, index


class HostMemory:
    """Classical program memory only available to the Host.
    Simple mapping from variable names to values."""

    def __init__(self, pid: int) -> None:
        self._pid = pid

        # Host memory is represented as a mapping from variables to values.
        # Variables have a name (str) and values (int).
        self._mem: Dict[str, int] = {}

    def write(self, loc: str, value: int) -> None:
        self._mem[loc] = value

    def read(self, loc: str) -> int:
        return self._mem[loc]


class SharedMemory:
    """Classical program memory available to both Host and Qnos.
    Implemented as NetQASM arrays and registers.

    TODO: registers should be moved to QnosMemory."""

    def __init__(self, pid: int) -> None:
        self._pid = pid

        register_names: Dict[RegisterName, RegisterGroup] = setup_registers()
        self._registers: Dict[Dict[RegisterName, RegisterGroup], int] = {}
        # TODO fix this abomination of handling registers
        for name in register_names.keys():
            self._registers[name] = {}  # type: ignore
            for i in range(16):
                self._registers[name][i] = 0  # type: ignore
        self._arrays: Arrays = Arrays()

    def set_reg_value(self, register: Union[str, operand.Register], value: int) -> None:
        if isinstance(register, str):
            name, index = RegisterMeta.parse(register)
        else:
            name, index = register.name, register.index
        self._registers[name][index] = value  # type: ignore

    def get_reg_value(self, register: Union[str, operand.Register]) -> int:
        if isinstance(register, str):
            name, index = RegisterMeta.parse(register)
        else:
            name, index = register.name, register.index
        return self._registers[name][index]  # type: ignore

    # for compatibility with netqasm Futures
    def get_register(self, register: Union[str, operand.Register]) -> Optional[int]:
        return self.get_reg_value(register)

    # for compatibility with netqasm Futures
    def get_array_part(
        self, address: int, index: Union[int, slice]
    ) -> Union[None, int, List[Optional[int]]]:
        if isinstance(index, int):
            return self.get_array_value(address, index)
        elif isinstance(index, slice):
            return self.get_array_values(address, index.start, index.stop)

    def init_new_array(self, address: int, length: int) -> None:
        self._arrays.init_new_array(address, length)

    def get_array(self, address: int) -> List[Optional[int]]:
        return self._arrays._get_array(address)  # type: ignore

    def get_array_entry(self, array_entry: operand.ArrayEntry) -> Optional[int]:
        address, index = self.expand_array_part(array_part=array_entry)
        result = self._arrays[address, index]
        assert (result is None) or isinstance(result, int)
        return result

    def get_array_value(self, addr: int, offset: int) -> Optional[int]:
        address, index = self.expand_array_part(
            array_part=operand.ArrayEntry(operand.Address(addr), offset)
        )
        result = self._arrays[address, index]
        assert (result is None) or isinstance(result, int)
        return result

    def get_array_values(
        self, addr: int, start_offset: int, end_offset
    ) -> List[Optional[int]]:
        values = self.get_array_slice(
            operand.ArraySlice(operand.Address(addr), start_offset, end_offset)
        )
        assert values is not None
        return values

    def set_array_entry(
        self, array_entry: operand.ArrayEntry, value: Optional[int]
    ) -> None:
        address, index = self.expand_array_part(array_part=array_entry)
        self._arrays[address, index] = value

    def set_array_value(self, addr: int, offset: int, value: Optional[int]) -> None:
        address, index = self.expand_array_part(
            array_part=operand.ArrayEntry(operand.Address(addr), offset)
        )
        self._arrays[address, index] = value

    def get_array_slice(
        self, array_slice: operand.ArraySlice
    ) -> Optional[List[Optional[int]]]:
        address, index = self.expand_array_part(array_part=array_slice)
        result = self._arrays[address, index]
        assert (result is None) or isinstance(result, list)
        return result

    def expand_array_part(
        self, array_part: Union[operand.ArrayEntry, operand.ArraySlice]
    ) -> Tuple[int, Union[int, slice]]:
        address: int = array_part.address.address
        index: Union[int, slice]
        if isinstance(array_part, operand.ArrayEntry):
            if isinstance(array_part.index, int):
                index = array_part.index
            else:
                index_from_reg = self.get_reg_value(register=array_part.index)
                if index_from_reg is None:
                    raise RuntimeError(
                        f"Trying to use register {array_part.index} "
                        "to index an array but its value is None"
                    )
                index = index_from_reg
        elif isinstance(array_part, operand.ArraySlice):
            startstop: List[int] = []
            for raw_s in [array_part.start, array_part.stop]:
                if isinstance(raw_s, int):
                    startstop.append(raw_s)
                elif isinstance(raw_s, operand.Register):
                    s = self.get_reg_value(register=raw_s)
                    if s is None:
                        raise RuntimeError(
                            f"Trying to use register {raw_s} to "
                            "index an array but its value is None"
                        )
                    startstop.append(s)
                else:
                    raise RuntimeError(
                        f"Something went wrong: raw_s should be int "
                        f"or Register but is {type(raw_s)}"
                    )
            index = slice(*startstop)
        else:
            raise RuntimeError(
                f"Something went wrong: array_part is a {type(array_part)}"
            )
        return address, index


class QnosMemory:
    """Classical program memory only available to Qnos.
    Not used at the moment.

    TODO: move NetQASM registers into here."""

    def __init__(self, pid: int) -> None:
        self._pid = pid


class QubitTrait:
    pass


class CommQubitTrait(QubitTrait):
    pass


class MemQubitTrait(QubitTrait):
    pass


class DecorenceQubitTrait(QubitTrait):
    def __init__(self, t1: int, t2: int) -> None:
        self._t1, self._t2 = t1, t2


class GateTrait:
    pass


class SingleGateTrait(GateTrait):
    def __init__(self, depolarizing_factor: float) -> None:
        self._depolarizing_factor = depolarizing_factor


class TwoGateTrait(GateTrait):
    def __init__(self, depolarizing_factor: float) -> None:
        self._depolarizing_factor = depolarizing_factor


@dataclass
class Topology:
    comm_ids: Set[int]
    mem_ids: Set[int]


@dataclass(eq=True, frozen=True)
class UnitModule:
    """
    :param qubit_ids: list of qubit IDs
    :param qubit_traits: map from qubit ID to list of qubit traits
    :param gate_traits: map from list of qubit IDs to list of gate traits
    """

    qubit_ids: List[int]
    qubit_traits: Dict[int, List[Type[QubitTrait]]]
    gate_traits: Dict[List[int], List[GateTrait]]

    @property
    def num_qubits(self) -> int:
        return len(self.qubit_ids)

    @classmethod
    def from_topology(cls, topology: Topology) -> UnitModule:
        all_ids = topology.comm_ids.union(topology.mem_ids)
        traits: Dict[int, List[Type[QubitTrait]]] = {i: [] for i in all_ids}
        for i in all_ids:
            if i in topology.comm_ids:
                traits[i].append(CommQubitTrait)
            if i in topology.mem_ids:
                traits[i].append(MemQubitTrait)
        return UnitModule(qubit_ids=list(all_ids), qubit_traits=traits, gate_traits={})

    @classmethod
    def default_generic(cls, num_qubits: int) -> UnitModule:
        return UnitModule(
            qubit_ids=[i for i in range(num_qubits)],
            qubit_traits={i: [CommQubitTrait, MemQubitTrait] for i in range(num_qubits)},  # type: ignore
            gate_traits={},
        )


class QuantumMemory:
    """Quantum memory only available to Qnos. Represented as unit modules."""

    # Only describes the virtual memory space (i.e. unit module).
    # Does not contain 'values' (quantum states) of the virtual memory locations.
    # Does not contain the mapping from virtual to physical space. (Managed by memmgr)

    def __init__(self, pid: int, unit_module: UnitModule) -> None:
        self._pid = pid
        self._unit_module = unit_module

    @property
    def unit_module(self) -> UnitModule:
        return self._unit_module


class ProgramMemory:
    def __init__(self, pid: int, unit_module: UnitModule) -> None:
        self._pid: int = pid

        self._host_memory = HostMemory(pid)
        self._shared_memory = SharedMemory(pid)
        self._qnos_memory = QnosMemory(pid)
        self._quantum_memory = QuantumMemory(pid, unit_module)

        self._prog_counter: int = 0

    @property
    def host_mem(self) -> HostMemory:
        return self._host_memory

    @property
    def shared_mem(self) -> SharedMemory:
        return self._shared_memory

    @property
    def qnos_mem(self) -> QnosMemory:
        return self._qnos_memory

    @property
    def quantum_mem(self) -> QuantumMemory:
        return self._quantum_memory

    @property
    def prog_counter(self) -> int:
        return self._prog_counter

    def increment_prog_counter(self) -> None:
        self._prog_counter += 1

    def set_prog_counter(self, value: int) -> None:
        self._prog_counter = value



import logging
from dataclasses import dataclass
from typing import Dict, Optional

from qoala.sim.hostprocessor import IqoalaProcess
from qoala.sim.logging import LogManager
from qoala.sim.memory import CommQubitTrait, MemQubitTrait, UnitModule
from qoala.sim.qdevice import QDevice


class AllocError(Exception):
    """Allocating a qubit did not succeed."""

    pass


class NotAllocatedError(Exception):
    """A virtual qubit is not mapped to any physical qubit."""

    pass


@dataclass
class VirtualMapping:
    # mapping from virt ID in specific unit module to phys ID
    unit_module: UnitModule
    mapping: Dict[int, Optional[int]]  # virt ID -> phys ID


@dataclass
class VirtualLocation:
    # particular virt ID in particular unit module in particular process
    pid: int
    unit_module: UnitModule
    virt_id: int


class MemoryManager:
    def __init__(self, node_name: str, qdevice: QDevice) -> None:
        self._node_name = node_name
        self._processes: Dict[int, IqoalaProcess] = {}
        self._logger: logging.Logger = LogManager.get_stack_logger(  # type: ignore
            f"{self.__class__.__name__}({self._node_name})"
        )

        self._qdevice = qdevice
        self._process_mappings: Dict[int, VirtualMapping] = {}  # pid -> mapping
        self._physical_mapping: Dict[int, Optional[VirtualLocation]] = {
            i: None for i in qdevice.all_qubit_ids
        }  # phys ID -> virt location

    def _virt_qubit_is_comm(self, unit_module: UnitModule, virt_id: int) -> bool:
        traits = unit_module.qubit_traits[virt_id]
        return CommQubitTrait in traits

    def _virt_qubit_is_mem(self, unit_module: UnitModule, virt_id: int) -> bool:
        traits = unit_module.qubit_traits[virt_id]
        return MemQubitTrait in traits

    def _get_free_comm_phys_id(self) -> int:
        for phys_id in self._qdevice.comm_qubit_ids:
            if self._physical_mapping[phys_id] is None:
                return phys_id  # type: ignore
        raise AllocError

    def _get_free_mem_phys_id(self) -> int:
        for phys_id in self._qdevice.mem_qubit_ids:
            if self._physical_mapping[phys_id] is None:
                return phys_id  # type: ignore
        raise AllocError

    def add_process(self, process: IqoalaProcess) -> None:
        self._processes[process.pid] = process
        unit_module = process.prog_memory.quantum_mem.unit_module
        self._process_mappings[process.pid] = VirtualMapping(
            unit_module, {x: None for x in unit_module.qubit_ids}
        )

    def get_process(self, pid: int) -> IqoalaProcess:
        return self._processes[pid]

    def allocate(self, pid: int, virt_id: int) -> int:
        vmap = self._process_mappings[pid]
        # Check if the virtual ID is in the unit module
        if virt_id not in vmap.unit_module.qubit_ids:
            raise AllocError

        # Check whether this virt ID is already mapped to a physical qubit.
        if vmap.mapping[virt_id] is not None:
            raise AllocError

        phys_id: int
        if self._virt_qubit_is_comm(vmap.unit_module, virt_id):
            phys_id = self._get_free_comm_phys_id()
        else:
            phys_id = self._get_free_mem_phys_id()

        # update mappings
        self._physical_mapping[phys_id] = VirtualLocation(
            pid, vmap.unit_module, virt_id
        )
        self._process_mappings[pid].mapping[virt_id] = phys_id
        return phys_id

    def allocate_comm(self, pid: int, virt_id: int) -> int:
        vmap = self._process_mappings[pid]
        # Check that the virt ID is indeed a (virtual) comm qubit.
        if virt_id not in vmap.unit_module.qubit_ids:
            raise AllocError
        if not self._virt_qubit_is_comm(vmap.unit_module, virt_id):
            raise AllocError

        return self.allocate(pid, virt_id)

    def free(self, pid: int, virt_id: int) -> None:
        vmap = self._process_mappings[pid]
        # Check if the virtual ID is in the unit module
        assert virt_id in vmap.unit_module.qubit_ids
        assert virt_id in self._process_mappings[pid].mapping

        phys_id = self._process_mappings[pid].mapping[virt_id]
        if phys_id is None:
            raise AllocError

        # update mappings
        self._physical_mapping[phys_id] = None
        self._process_mappings[pid].mapping[virt_id] = None

        # update netsquid memory
        self._qdevice.set_mem_pos_in_use(phys_id, False)

    def get_unmapped_mem_qubit(self, pid: int) -> int:
        """returns virt ID"""
        vp_map = self._process_mappings[pid].mapping
        unit_module = self._process_mappings[pid].unit_module
        free_ids = [
            v
            for v, p in vp_map.items()
            if p is None and self._virt_qubit_is_mem(unit_module, v)
        ]
        if len(free_ids) == 0:
            raise AllocError
        return min(free_ids)

    def phys_id_for(self, pid: int, virt_id: int) -> Optional[int]:
        phys_id = self._process_mappings[pid].mapping[virt_id]
        return phys_id

    def virt_id_for(self, pid: int, phys_id: int) -> Optional[int]:
        if virt_loc := self._physical_mapping[phys_id]:
            if virt_loc.pid == pid:
                return virt_loc.virt_id
        return None



import itertools
from typing import Dict, List

import numpy as np
from netsquid.components.instructions import (
    INSTR_CNOT,
    INSTR_CXDIR,
    INSTR_CYDIR,
    INSTR_CZ,
    INSTR_H,
    INSTR_INIT,
    INSTR_MEASURE,
    INSTR_ROT_X,
    INSTR_ROT_Y,
    INSTR_ROT_Z,
    INSTR_X,
    INSTR_Y,
    INSTR_Z,
)
from netsquid.components.models.qerrormodels import DepolarNoiseModel, T1T2NoiseModel
from netsquid.components.qprocessor import PhysicalInstruction, QuantumProcessor
from netsquid.qubits.operators import Operator
from netsquid_magic.link_layer import (
    MagicLinkLayerProtocol,
    MagicLinkLayerProtocolWithSignaling,
    SingleClickTranslationUnit,
)
from netsquid_magic.magic_distributor import (
    DepolariseWithFailureMagicDistributor,
    DoubleClickMagicDistributor,
    PerfectStateMagicDistributor,
)
from netsquid_nv.magic_distributor import NVSingleClickMagicDistributor
from netsquid_physlayer.heralded_connection import MiddleHeraldedConnection

from qoala.runtime.config import (
    DepolariseLinkConfig,
    GenericQDeviceConfig,
    HeraldedLinkConfig,
    LinkConfig,
    NVLinkConfig,
    NVQDeviceConfig,
    ProcNodeConfig,
    ProcNodeNetworkConfig,
)
from qoala.runtime.environment import GlobalEnvironment
from qoala.sim.network import ProcNodeNetwork
from qoala.sim.procnode import ProcNode


# TODO: move this to somewhere else
def fidelity_to_prob_max_mixed(fid: float) -> float:
    return (1 - fid) * 4.0 / 3.0


def build_generic_qprocessor(name: str, cfg: GenericQDeviceConfig) -> QuantumProcessor:
    phys_instructions = []

    single_qubit_gate_noise = DepolarNoiseModel(
        depolar_rate=cfg.single_qubit_gate_depolar_prob, time_independent=True
    )

    two_qubit_gate_noise = DepolarNoiseModel(
        depolar_rate=cfg.two_qubit_gate_depolar_prob, time_independent=True
    )

    phys_instructions.append(
        PhysicalInstruction(
            INSTR_INIT,
            parallel=False,
            duration=cfg.init_time,
        )
    )

    for instr in [
        INSTR_ROT_X,
        INSTR_ROT_Y,
        INSTR_ROT_Z,
        INSTR_X,
        INSTR_Y,
        INSTR_Z,
        INSTR_H,
    ]:
        phys_instructions.append(
            PhysicalInstruction(
                instr,
                parallel=False,
                quantum_noise_model=single_qubit_gate_noise,
                apply_q_noise_after=True,
                duration=cfg.single_qubit_gate_time,
            )
        )

    for instr in [INSTR_CNOT, INSTR_CZ]:
        phys_instructions.append(
            PhysicalInstruction(
                instr,
                parallel=False,
                quantum_noise_model=two_qubit_gate_noise,
                apply_q_noise_after=True,
                duration=cfg.two_qubit_gate_time,
            )
        )

    phys_instr_measure = PhysicalInstruction(
        INSTR_MEASURE,
        parallel=False,
        duration=cfg.measure_time,
    )
    phys_instructions.append(phys_instr_measure)

    electron_qubit_noise = T1T2NoiseModel(T1=cfg.T1, T2=cfg.T2)
    mem_noise_models = [electron_qubit_noise] * cfg.num_qubits
    qmem = QuantumProcessor(
        name=name,
        num_positions=cfg.num_qubits,
        mem_noise_models=mem_noise_models,
        phys_instructions=phys_instructions,
    )
    return qmem


def build_nv_qprocessor(name: str, cfg: NVQDeviceConfig) -> QuantumProcessor:

    # noise models for single- and multi-qubit operations
    electron_init_noise = DepolarNoiseModel(
        depolar_rate=cfg.electron_init_depolar_prob, time_independent=True
    )

    electron_single_qubit_noise = DepolarNoiseModel(
        depolar_rate=cfg.electron_single_qubit_depolar_prob, time_independent=True
    )

    carbon_init_noise = DepolarNoiseModel(
        depolar_rate=cfg.carbon_init_depolar_prob, time_independent=True
    )

    carbon_z_rot_noise = DepolarNoiseModel(
        depolar_rate=cfg.carbon_z_rot_depolar_prob, time_independent=True
    )

    ec_noise = DepolarNoiseModel(
        depolar_rate=cfg.ec_gate_depolar_prob, time_independent=True
    )

    electron_qubit_noise = T1T2NoiseModel(T1=cfg.electron_T1, T2=cfg.electron_T2)

    carbon_qubit_noise = T1T2NoiseModel(T1=cfg.carbon_T1, T2=cfg.carbon_T2)

    # defining gates and their gate times

    phys_instructions = []

    electron_position = 0
    carbon_positions = [pos + 1 for pos in range(cfg.num_qubits - 1)]

    phys_instructions.append(
        PhysicalInstruction(
            INSTR_INIT,
            parallel=False,
            topology=carbon_positions,
            quantum_noise_model=carbon_init_noise,
            apply_q_noise_after=True,
            duration=cfg.carbon_init,
        )
    )

    for (instr, dur) in zip(
        [INSTR_ROT_X, INSTR_ROT_Y, INSTR_ROT_Z],
        [cfg.carbon_rot_x, cfg.carbon_rot_y, cfg.carbon_rot_z],
    ):
        phys_instructions.append(
            PhysicalInstruction(
                instr,
                parallel=False,
                topology=carbon_positions,
                quantum_noise_model=carbon_z_rot_noise,
                apply_q_noise_after=True,
                duration=dur,
            )
        )

    phys_instructions.append(
        PhysicalInstruction(
            INSTR_INIT,
            parallel=False,
            topology=[electron_position],
            quantum_noise_model=electron_init_noise,
            apply_q_noise_after=True,
            duration=cfg.electron_init,
        )
    )

    for (instr, dur) in zip(
        [INSTR_ROT_X, INSTR_ROT_Y, INSTR_ROT_Z],
        [cfg.electron_rot_x, cfg.electron_rot_y, cfg.electron_rot_z],
    ):
        phys_instructions.append(
            PhysicalInstruction(
                instr,
                parallel=False,
                topology=[electron_position],
                quantum_noise_model=electron_single_qubit_noise,
                apply_q_noise_after=True,
                duration=dur,
            )
        )

    electron_carbon_topologies = [
        (electron_position, carbon_pos) for carbon_pos in carbon_positions
    ]
    phys_instructions.append(
        PhysicalInstruction(
            INSTR_CXDIR,
            parallel=False,
            topology=electron_carbon_topologies,
            quantum_noise_model=ec_noise,
            apply_q_noise_after=True,
            duration=cfg.ec_controlled_dir_x,
        )
    )

    phys_instructions.append(
        PhysicalInstruction(
            INSTR_CYDIR,
            parallel=False,
            topology=electron_carbon_topologies,
            quantum_noise_model=ec_noise,
            apply_q_noise_after=True,
            duration=cfg.ec_controlled_dir_y,
        )
    )

    M0 = Operator(
        "M0", np.diag([np.sqrt(1 - cfg.prob_error_0), np.sqrt(cfg.prob_error_1)])
    )
    M1 = Operator(
        "M1", np.diag([np.sqrt(cfg.prob_error_0), np.sqrt(1 - cfg.prob_error_1)])
    )

    # hack to set imperfect measurements
    INSTR_MEASURE._meas_operators = [M0, M1]

    phys_instr_measure = PhysicalInstruction(
        INSTR_MEASURE,
        parallel=False,
        topology=[electron_position],
        quantum_noise_model=None,
        duration=cfg.measure,
    )

    phys_instructions.append(phys_instr_measure)

    # add qubits
    mem_noise_models = [electron_qubit_noise] + [carbon_qubit_noise] * len(
        carbon_positions
    )
    qmem = QuantumProcessor(
        name=name,
        num_positions=cfg.num_qubits,
        mem_noise_models=mem_noise_models,
        phys_instructions=phys_instructions,
    )
    return qmem


def build_procnode(cfg: ProcNodeConfig, global_env: GlobalEnvironment) -> ProcNode:
    if cfg.qdevice_typ == "nv":
        qdevice_cfg = cfg.qdevice_cfg
        if not isinstance(qdevice_cfg, NVQDeviceConfig):
            qdevice_cfg = NVQDeviceConfig(**cfg.qdevice_cfg)
        qprocessor = build_nv_qprocessor(f"qdevice_{cfg.name}", cfg=qdevice_cfg)
        procnode = ProcNode(
            cfg.name,
            global_env=global_env,
            qprocessor=qprocessor,
            qdevice_type="nv",
            node_id=cfg.node_id,
        )
    elif cfg.qdevice_typ == "generic":
        qdevice_cfg = cfg.qdevice_cfg
        if not isinstance(qdevice_cfg, GenericQDeviceConfig):
            qdevice_cfg = GenericQDeviceConfig(**cfg.qdevice_cfg)
        qprocessor = build_generic_qprocessor(f"qdevice_{cfg.name}", cfg=qdevice_cfg)
        procnode = ProcNode(
            cfg.name,
            global_env=global_env,
            qprocessor=qprocessor,
            qdevice_type="generic",
            node_id=cfg.node_id,
        )
        # TODO: do this in constructor?
        procnode.qnos.processor.instr_latency = cfg.instr_latency
        procnode.host.processor.instr_latency = cfg.instr_latency
        procnode.host.processor.receive_latency = cfg.receive_latency
    return procnode


def build_ll_protocol(
    config: LinkConfig, proc_node1: ProcNode, proc_node2: ProcNode
) -> MagicLinkLayerProtocolWithSignaling:
    if config.typ == "perfect":
        link_dist = PerfectStateMagicDistributor(
            nodes=[proc_node1.node, proc_node2.node], state_delay=0
        )
    elif config.typ == "depolarise":
        link_cfg = config.cfg
        if not isinstance(link_cfg, DepolariseLinkConfig):
            link_cfg = DepolariseLinkConfig(**config.cfg)
        prob_max_mixed = fidelity_to_prob_max_mixed(link_cfg.fidelity)
        link_dist = DepolariseWithFailureMagicDistributor(
            nodes=[proc_node1.node, proc_node2.node],
            prob_max_mixed=prob_max_mixed,
            prob_success=link_cfg.prob_success,
            t_cycle=link_cfg.t_cycle,
        )
    elif config.typ == "nv":
        link_cfg = config.cfg
        if not isinstance(link_cfg, NVLinkConfig):
            link_cfg = NVLinkConfig(**config.cfg)
        link_dist = NVSingleClickMagicDistributor(
            nodes=[proc_node1.node, proc_node2.node],
            length_A=link_cfg.length_A,
            length_B=link_cfg.length_B,
            full_cycle=link_cfg.full_cycle,
            cycle_time=link_cfg.cycle_time,
            alpha=link_cfg.alpha,
        )
    elif config.typ == "heralded":
        link_cfg = config.cfg
        if not isinstance(link_cfg, HeraldedLinkConfig):
            link_cfg = HeraldedLinkConfig(**config.cfg)
        connection = MiddleHeraldedConnection(name="heralded_conn", **link_cfg.dict())
        link_dist = DoubleClickMagicDistributor(
            [proc_node1.node, proc_node2.node], connection
        )
    else:
        raise ValueError

    return MagicLinkLayerProtocolWithSignaling(
        nodes=[proc_node1.node, proc_node2.node],
        magic_distributor=link_dist,
        translation_unit=SingleClickTranslationUnit(),
    )


def build_network(
    config: ProcNodeNetworkConfig, global_env: GlobalEnvironment
) -> ProcNodeNetwork:
    proc_nodes: Dict[str, ProcNode] = {}
    link_prots: List[MagicLinkLayerProtocol] = []

    for cfg in config.nodes:
        proc_nodes[cfg.name] = build_procnode(cfg, global_env)

    for (_, s1), (_, s2) in itertools.combinations(proc_nodes.items(), 2):
        s1.connect_to(s2)

    for cfg in config.links:
        proc_node1 = proc_nodes[cfg.node1]
        proc_node2 = proc_nodes[cfg.node2]
        link_prot = build_ll_protocol(cfg, proc_node1, proc_node2)
        proc_node1.assign_ll_protocol(proc_node2.node.ID, link_prot)
        proc_node2.assign_ll_protocol(proc_node1.node.ID, link_prot)

        link_prots.append(link_prot)

    return ProcNodeNetwork(proc_nodes, link_prots)



import os

import pytest
from netqasm.lang.instr.core import MeasInstruction, SetInstruction
from netqasm.lang.operand import Register, Template
from netqasm.lang.subroutine import Subroutine

from qoala.lang.iqoala import (
    AssignCValueOp,
    IqoalaInstrParser,
    IqoalaMetaParser,
    IqoalaParseError,
    IqoalaParser,
    IqoalaRequest,
    IQoalaRequestParser,
    IqoalaSharedMemLoc,
    IqoalaSubroutine,
    IQoalaSubroutineParser,
    IqoalaVector,
    ProgramMeta,
    RunSubroutineOp,
)
from qoala.sim.requests import (
    EprCreateRole,
    EprCreateType,
    NetstackCreateRequest,
    NetstackReceiveRequest,
)
from qoala.util.tests import text_equal


def test_parse_incomplete_meta():
    text = """
META_START
name: alice
parameters: 
csockets: 0 -> bob
META_END
    """

    with pytest.raises(IqoalaParseError):
        IqoalaMetaParser(text).parse()


def test_parse_meta_no_end():
    text = """
META_START
name: alice
parameters: 
csockets: 0 -> bob
epr_sockets: 
    """

    with pytest.raises(IqoalaParseError):
        IqoalaMetaParser(text).parse()


def test_parse_meta():
    text = """
META_START
name: alice
parameters: 
csockets: 0 -> bob
epr_sockets: 
META_END
    """

    meta = IqoalaMetaParser(text).parse()

    assert meta == ProgramMeta(
        name="alice", parameters=[], csockets={0: "bob"}, epr_sockets={}
    )


def test_parse_meta_multiple_remotes():
    text = """
META_START
name: alice
parameters: theta1, theta2
csockets: 0 -> bob, 1 -> charlie
epr_sockets: 0 -> bob
META_END
    """

    meta = IqoalaMetaParser(text).parse()

    assert meta == ProgramMeta(
        name="alice",
        parameters=["theta1", "theta2"],
        csockets={0: "bob", 1: "charlie"},
        epr_sockets={0: "bob"},
    )


def test_parse_1_instr():
    text = """
x = assign_cval() : 1
    """

    instructions = IqoalaInstrParser(text).parse()

    assert len(instructions) == 1
    assert instructions[0] == AssignCValueOp(result="x", value=1)


def test_parse_2_instr():
    text = """
x = assign_cval() : 1
y = assign_cval() : 17
    """

    instructions = IqoalaInstrParser(text).parse()

    assert len(instructions) == 2
    assert instructions[0] == AssignCValueOp(result="x", value=1)
    assert instructions[1] == AssignCValueOp(result="y", value=17)


def test_parse_faulty_instr():
    text = """
x = assign_cval
    """

    with pytest.raises(IqoalaParseError):
        IqoalaInstrParser(text).parse()


def test_parse_vec():
    text = """
run_subroutine(vec<x>) : subrt1
    """

    instructions = IqoalaInstrParser(text).parse()
    assert len(instructions) == 1
    assert instructions[0] == RunSubroutineOp(
        result=None, values=IqoalaVector(["x"]), subrt="subrt1"
    )


def test_parse_vec_2_elements():
    text = """
run_subroutine(vec<x; y>) : subrt1
    """

    instructions = IqoalaInstrParser(text).parse()
    assert len(instructions) == 1
    assert instructions[0] == RunSubroutineOp(
        result=None, values=IqoalaVector(["x", "y"]), subrt="subrt1"
    )


def test_parse_vec_2_elements_and_return():
    text = """
vec<m> = run_subroutine(vec<x; y>) : subrt1
    """

    instructions = IqoalaInstrParser(text).parse()
    assert len(instructions) == 1
    assert instructions[0] == RunSubroutineOp(
        result=IqoalaVector(["m"]), values=IqoalaVector(["x", "y"]), subrt="subrt1"
    )


def test_parse_vec_2_elements_and_return_2_elements():
    text = """
vec<m1; m2> = run_subroutine(vec<x; y>) : subrt1
    """

    instructions = IqoalaInstrParser(text).parse()
    assert len(instructions) == 1
    assert instructions[0] == RunSubroutineOp(
        result=IqoalaVector(["m1", "m2"]),
        values=IqoalaVector(["x", "y"]),
        subrt="subrt1",
    )


def test_parse_subrt():
    text = """
SUBROUTINE subrt1
    params: my_value
    returns: M0 -> m
    request: 
  NETQASM_START
    set Q0 {my_value}
  NETQASM_END
    """

    parsed = IQoalaSubroutineParser(text).parse()
    assert len(parsed) == 1
    assert "subrt1" in parsed
    subrt = parsed["subrt1"]

    expected_instrs = [
        SetInstruction(reg=Register.from_str("Q0"), imm=Template("my_value"))
    ]
    expected_args = ["my_value"]
    assert subrt == IqoalaSubroutine(
        name="subrt1",
        subrt=Subroutine(instructions=expected_instrs, arguments=expected_args),
        return_map={"m": IqoalaSharedMemLoc("M0")},
    )


def test_parse_subrt_2():
    text = """
SUBROUTINE my_subroutine
    params: param1, param2
    returns: M5 -> result1, M6 -> result2
    request: 
  NETQASM_START
    set R0 {param1}
    set R1 {param2}
    meas Q0 M5
    meas Q1 M6
  NETQASM_END
    """

    parsed = IQoalaSubroutineParser(text).parse()
    assert len(parsed) == 1
    assert "my_subroutine" in parsed
    subrt = parsed["my_subroutine"]

    expected_instrs = [
        SetInstruction(reg=Register.from_str("R0"), imm=Template("param1")),
        SetInstruction(reg=Register.from_str("R1"), imm=Template("param2")),
        MeasInstruction(reg0=Register.from_str("Q0"), reg1=Register.from_str("M5")),
        MeasInstruction(reg0=Register.from_str("Q1"), reg1=Register.from_str("M6")),
    ]
    expected_args = ["param1", "param2"]
    assert subrt == IqoalaSubroutine(
        name="my_subroutine",
        subrt=Subroutine(instructions=expected_instrs, arguments=expected_args),
        return_map={
            "result1": IqoalaSharedMemLoc("M5"),
            "result2": IqoalaSharedMemLoc("M6"),
        },
    )


def test_parse_multiple_subrt():
    text = """
SUBROUTINE subrt1
    params: param1
    returns: M0 -> m
    request: 
  NETQASM_START
    set R0 {param1}
    meas Q0 M0
  NETQASM_END

SUBROUTINE subrt2
    params: theta
    returns: 
    request: 
  NETQASM_START
    set R0 {theta}
  NETQASM_END
    """

    parsed = IQoalaSubroutineParser(text).parse()
    assert len(parsed) == 2
    assert "subrt1" in parsed
    assert "subrt2" in parsed
    subrt1 = parsed["subrt1"]
    subrt2 = parsed["subrt2"]

    expected_instrs_1 = [
        SetInstruction(reg=Register.from_str("R0"), imm=Template("param1")),
        MeasInstruction(reg0=Register.from_str("Q0"), reg1=Register.from_str("M0")),
    ]
    expected_args_1 = ["param1"]
    expected_instrs_2 = [
        SetInstruction(reg=Register.from_str("R0"), imm=Template("theta")),
    ]
    expected_args_2 = ["theta"]

    assert subrt1 == IqoalaSubroutine(
        name="subrt1",
        subrt=Subroutine(instructions=expected_instrs_1, arguments=expected_args_1),
        return_map={
            "m": IqoalaSharedMemLoc("M0"),
        },
    )
    assert subrt2 == IqoalaSubroutine(
        name="subrt2",
        subrt=Subroutine(instructions=expected_instrs_2, arguments=expected_args_2),
        return_map={},
    )


def test_parse_invalid_subrt():
    text = """
SUBROUTINE my_subroutine
    params: param1, param2
    returns: M5 -> result1, M6 -> result2
    request: 
  NETQASM_START
    set R0 {param3}
  NETQASM_END
    """

    with pytest.raises(IqoalaParseError):
        IQoalaSubroutineParser(text).parse()


def test_parse_request():
    text = """
REQUEST req1
  role: create
  remote_id: 1
  epr_socket_id: 0
  typ: create_keep
  num_pairs: 5
  fidelity: 0.65
  virt_qubit_ids: 0, 0, 0, 0, 0
  result_array_addr: 3
    """

    parsed = IQoalaRequestParser(text).parse()
    assert len(parsed) == 1
    assert "req1" in parsed
    request = parsed["req1"]

    assert request == IqoalaRequest(
        name="req1",
        role=EprCreateRole.CREATE,
        request=NetstackCreateRequest(
            remote_id=1,
            epr_socket_id=0,
            typ=EprCreateType.CREATE_KEEP,
            num_pairs=5,
            fidelity=0.65,
            virt_qubit_ids=[0, 0, 0, 0, 0],
            result_array_addr=3,
        ),
    )


def test_parse_request_2():
    text = """
REQUEST req1
  role: receive
  remote_id: 1
  epr_socket_id: 0
  typ: measure_directly
  num_pairs: 3
  fidelity: 0.65
  virt_qubit_ids: 1, 2, 3
  result_array_addr: 0
    """

    parsed = IQoalaRequestParser(text).parse()
    assert len(parsed) == 1
    assert "req1" in parsed
    request = parsed["req1"]

    assert request == IqoalaRequest(
        name="req1",
        role=EprCreateRole.RECEIVE,
        request=NetstackReceiveRequest(
            remote_id=1,
            epr_socket_id=0,
            typ=EprCreateType.MEASURE_DIRECTLY,
            num_pairs=3,
            fidelity=0.65,
            virt_qubit_ids=[1, 2, 3],
            result_array_addr=0,
        ),
    )


def test_parse_request_with_template():
    text = """
REQUEST req1
  role: receive
  remote_id: {client_id}
  epr_socket_id: 0
  typ: measure_directly
  num_pairs: 3
  fidelity: 0.65
  virt_qubit_ids: 1, 2, 3
  result_array_addr: 0
    """

    parsed = IQoalaRequestParser(text).parse()
    assert len(parsed) == 1
    assert "req1" in parsed
    request = parsed["req1"]

    assert request == IqoalaRequest(
        name="req1",
        role=EprCreateRole.RECEIVE,
        request=NetstackReceiveRequest(
            remote_id=Template("client_id"),
            epr_socket_id=0,
            typ=EprCreateType.MEASURE_DIRECTLY,
            num_pairs=3,
            fidelity=0.65,
            virt_qubit_ids=[1, 2, 3],
            result_array_addr=0,
        ),
    )


def test_parse_multiple_request():
    text = """
REQUEST req1
  role: create
  remote_id: 1
  epr_socket_id: 0
  typ: create_keep
  num_pairs: 5
  fidelity: 0.65
  virt_qubit_ids: 0, 0, 0, 0, 0
  result_array_addr: 3

REQUEST req2
  role: receive
  remote_id: 1
  epr_socket_id: 0
  typ: measure_directly
  num_pairs: 3
  fidelity: 0.65
  virt_qubit_ids: 1, 2, 3
  result_array_addr: 0
    """

    parsed = IQoalaRequestParser(text).parse()
    assert len(parsed) == 2
    assert "req1" in parsed
    assert "req2" in parsed
    req1 = parsed["req1"]
    req2 = parsed["req2"]

    assert req1 == IqoalaRequest(
        name="req1",
        role=EprCreateRole.CREATE,
        request=NetstackCreateRequest(
            remote_id=1,
            epr_socket_id=0,
            typ=EprCreateType.CREATE_KEEP,
            num_pairs=5,
            fidelity=0.65,
            virt_qubit_ids=[0, 0, 0, 0, 0],
            result_array_addr=3,
        ),
    )
    assert req2 == IqoalaRequest(
        name="req2",
        role=EprCreateRole.RECEIVE,
        request=NetstackReceiveRequest(
            remote_id=1,
            epr_socket_id=0,
            typ=EprCreateType.MEASURE_DIRECTLY,
            num_pairs=3,
            fidelity=0.65,
            virt_qubit_ids=[1, 2, 3],
            result_array_addr=0,
        ),
    )


def test_parse_invalid_request():
    text = """
REQUEST req1
  role: receive
  remote_id: 1
  epr_socket_id: 0
  typ: measure
  num_pairs: 3
  fidelity: 0.65
  virt_qubit_ids: 1, 2, 3
  result_array_addr: 0
    """

    # invalid 'typ' value

    with pytest.raises(IqoalaParseError):
        IQoalaRequestParser(text).parse()


DEFAULT_META = """
META_START
name: alice
parameters: 
csockets: 0 -> bob
epr_sockets: 
META_END
"""


def test_parse_program():
    meta_text = """
META_START
name: alice
parameters: 
csockets: 0 -> bob
epr_sockets: 
META_END
    """

    program_text = """
my_value = assign_cval() : 1
remote_id = assign_cval() : 0
send_cmsg(remote_id, my_value)
received_value = recv_cmsg(remote_id)
new_value = assign_cval() : 3
my_value = add_cval_c(new_value, new_value)
vec<m> = run_subroutine(vec<my_value>) : subrt1
return_result(m)
    """

    subrt_text = """
SUBROUTINE subrt1
    params: my_value
    returns: M0 -> m
    request: 
  NETQASM_START
    set Q0 0
    rot_z Q0 {my_value} 4
    meas Q0 M0
    ret_reg M0
  NETQASM_END
    """

    req_text = """
REQUEST req1
  role: receive
  remote_id: 1
  epr_socket_id: 0
  typ: measure_directly
  num_pairs: 3
  fidelity: 0.65
  virt_qubit_ids: 1, 2, 3
  result_array_addr: 0
    """

    parsed_program = IqoalaParser(
        meta_text=meta_text,
        instr_text=program_text,
        subrt_text=subrt_text,
        req_text=req_text,
    ).parse()
    assert len(parsed_program.instructions) == 8
    assert "subrt1" in parsed_program.subroutines
    assert len(parsed_program.requests) == 1
    assert "req1" in parsed_program.requests


def test_parse_program_no_requests():
    meta_text = """
META_START
name: alice
parameters: 
csockets: 0 -> bob
epr_sockets: 
META_END
    """

    program_text = """
my_value = assign_cval() : 1
remote_id = assign_cval() : 0
send_cmsg(remote_id, my_value)
received_value = recv_cmsg(remote_id)
new_value = assign_cval() : 3
my_value = add_cval_c(new_value, new_value)
vec<m> = run_subroutine(vec<my_value>) : subrt1
return_result(m)
    """

    subrt_text = """
SUBROUTINE subrt1
    params: my_value
    returns: M0 -> m
    request: 
  NETQASM_START
    set Q0 0
    rot_z Q0 {my_value} 4
    meas Q0 M0
    ret_reg M0
  NETQASM_END
    """

    req_text = """
    """

    parsed_program = IqoalaParser(
        meta_text=meta_text,
        instr_text=program_text,
        subrt_text=subrt_text,
        req_text=req_text,
    ).parse()
    assert len(parsed_program.instructions) == 8
    assert "subrt1" in parsed_program.subroutines
    assert len(parsed_program.requests) == 0


def test_parse_program_invalid_subrt_reference():
    meta_text = """
META_START
name: alice
parameters: 
csockets: 0 -> bob
epr_sockets: 
META_END
    """

    program_text = """
my_value = assign_cval() : 1
vec<m> = run_subroutine(vec<my_value>) : non_existing_subrt
return_result(m)
    """

    subrt_text = """
SUBROUTINE subrt1
    params: my_value
    returns: M0 -> m
    request: 
  NETQASM_START
    set R0 0
  NETQASM_END
    """

    with pytest.raises(IqoalaParseError):
        IqoalaParser(
            meta_text=meta_text,
            instr_text=program_text,
            subrt_text=subrt_text,
            req_text="",
        ).parse()


def test_split_text():
    meta_text = """
META_START
name: alice
parameters: 
csockets: 0 -> bob
epr_sockets: 
META_END
    """

    instr_text = """
m = assign_cval() : 1
return_result(m)
    """

    subrt_text = """
SUBROUTINE subrt1
    params: 
    returns: 
    request: 
  NETQASM_START
    set Q0 0
  NETQASM_END
    """

    text = meta_text + instr_text + subrt_text
    parser = IqoalaParser(text)

    assert text_equal(parser._meta_text, meta_text)
    assert text_equal(parser._instr_text, instr_text)
    assert text_equal(parser._subrt_text, subrt_text)


def test_split_text_multiple_subroutines():
    meta_text = """
META_START
name: alice
parameters: 
csockets: 0 -> bob
epr_sockets: 
META_END
    """

    instr_text = """
m = assign_cval() : 1
return_result(m)
    """

    subrt_text = """
SUBROUTINE subrt1
    params: 
    returns: 
    request: 
  NETQASM_START
    set Q0 0
  NETQASM_END

SUBROUTINE subrt2
    params: 
    returns: 
    request: 
  NETQASM_START
    set Q7 7
  NETQASM_END
    """

    text = meta_text + instr_text + subrt_text
    parser = IqoalaParser(text)

    assert text_equal(parser._meta_text, meta_text)
    assert text_equal(parser._instr_text, instr_text)
    assert text_equal(parser._subrt_text, subrt_text)


def test_parse_program_single_text():
    text = """
META_START
name: alice
parameters: 
csockets: 0 -> bob
epr_sockets: 
META_END

my_value = assign_cval() : 1
remote_id = assign_cval() : 0
send_cmsg(remote_id, my_value)
received_value = recv_cmsg(remote_id)
new_value = assign_cval() : 3
my_value = add_cval_c(new_value, new_value)
vec<m> = run_subroutine(vec<my_value>) : subrt1
return_result(m)

SUBROUTINE subrt1
    params: my_value
    returns: M0 -> m
    request: 
  NETQASM_START
    set Q0 0
    rot_z Q0 {my_value} 4
    meas Q0 M0
    ret_reg M0
  NETQASM_END
    """

    parsed_program = IqoalaParser(text).parse()
    assert len(parsed_program.instructions) == 8
    assert "subrt1" in parsed_program.subroutines


def test_parse_file():
    path = os.path.join(os.path.dirname(__file__), "test_bqc_server.iqoala")
    with open(path) as file:
        text = file.read()
    parsed_program = IqoalaParser(text).parse()
    assert len(parsed_program.instructions) == 11
    assert "create_epr_0" in parsed_program.subroutines
    assert "create_epr_1" in parsed_program.subroutines
    assert "local_cphase" in parsed_program.subroutines
    assert "meas_qubit_0" in parsed_program.subroutines
    assert "meas_qubit_1" in parsed_program.subroutines
    assert "req0" in parsed_program.requests
    assert "req1" in parsed_program.requests
    assert parsed_program.subroutines["create_epr_0"].request_name == "req0"
    assert parsed_program.subroutines["create_epr_1"].request_name == "req1"


def test_parse_file_2():
    path = os.path.join(os.path.dirname(__file__), "test_bqc_client.iqoala")
    with open(path) as file:
        text = file.read()
    parsed_program = IqoalaParser(text).parse()
    assert len(parsed_program.instructions) == 19
    assert "create_epr_0" in parsed_program.subroutines
    assert "post_epr_0" in parsed_program.subroutines
    assert "create_epr_1" in parsed_program.subroutines
    assert "post_epr_1" in parsed_program.subroutines
    assert "req0" in parsed_program.requests
    assert "req1" in parsed_program.requests
    assert parsed_program.subroutines["create_epr_0"].request_name == "req0"
    assert parsed_program.subroutines["create_epr_1"].request_name == "req1"


if __name__ == "__main__":
    test_parse_incomplete_meta()
    test_parse_meta_no_end()
    test_parse_meta()
    test_parse_meta_multiple_remotes()
    test_parse_1_instr()
    test_parse_2_instr()
    test_parse_faulty_instr()
    test_parse_vec()
    test_parse_vec_2_elements()
    test_parse_vec_2_elements_and_return()
    test_parse_vec_2_elements_and_return_2_elements()
    test_parse_subrt()
    test_parse_subrt_2()
    test_parse_multiple_subrt()
    test_parse_invalid_subrt()
    test_parse_request()
    test_parse_request_2()
    test_parse_request_with_template()
    test_parse_multiple_request()
    test_parse_invalid_request()
    test_parse_program()
    test_parse_program_no_requests()
    test_parse_program_invalid_subrt_reference()
    test_split_text()
    test_split_text_multiple_subroutines()
    test_parse_program_single_text()
    test_parse_file()
    test_parse_file_2()



import pytest
from netsquid.components.instructions import (
    INSTR_CNOT,
    INSTR_CXDIR,
    INSTR_INIT,
    INSTR_MEASURE,
    INSTR_ROT_X,
    INSTR_X,
)
from netsquid.components.qprocessor import MissingInstructionError, QuantumProcessor

from qoala.runtime.config import GenericQDeviceConfig, NVQDeviceConfig
from qoala.sim.build import build_generic_qprocessor, build_nv_qprocessor


def test_build_generic_perfect():
    num_qubits = 2
    cfg = GenericQDeviceConfig.perfect_config(num_qubits)
    proc: QuantumProcessor = build_generic_qprocessor(name="alice", cfg=cfg)
    assert proc.num_positions == num_qubits

    for i in range(num_qubits):
        assert proc.get_instruction_duration(INSTR_INIT, [i]) == cfg.init_time
        assert proc.get_instruction_duration(INSTR_MEASURE, [i]) == cfg.measure_time
        assert proc.get_instruction_duration(INSTR_X, [i]) == cfg.single_qubit_gate_time
        assert (
            proc.get_instruction_duration(INSTR_ROT_X, [i])
            == cfg.single_qubit_gate_time
        )

    assert proc.get_instruction_duration(INSTR_CNOT, [0, 1]) == cfg.two_qubit_gate_time

    # TODO: check topology??!


def test_build_nv_perfect():
    num_qubits = 2
    cfg = NVQDeviceConfig.perfect_config(num_qubits)
    proc: QuantumProcessor = build_nv_qprocessor(name="alice", cfg=cfg)
    assert proc.num_positions == num_qubits

    assert proc.get_instruction_duration(INSTR_INIT, [0]) == cfg.electron_init
    assert proc.get_instruction_duration(INSTR_MEASURE, [0]) == cfg.measure
    assert proc.get_instruction_duration(INSTR_ROT_X, [0]) == cfg.electron_rot_x

    for i in range(1, num_qubits):
        assert proc.get_instruction_duration(INSTR_INIT, [i]) == cfg.carbon_init
        assert proc.get_instruction_duration(INSTR_ROT_X, [i]) == cfg.carbon_rot_x
        with pytest.raises(MissingInstructionError):
            assert proc.get_instruction_duration(INSTR_MEASURE, [i])

    with pytest.raises(MissingInstructionError):
        proc.get_instruction_duration(INSTR_CNOT, [0, 1])
        proc.get_instruction_duration(INSTR_CXDIR, [1, 0])

    assert proc.get_instruction_duration(INSTR_CXDIR, [0, 1]) == cfg.ec_controlled_dir_x


if __name__ == "__main__":
    test_build_generic_perfect()
    test_build_nv_perfect()



from __future__ import annotations

from typing import Dict

from netsquid.components.component import Component, Port
from netsquid.nodes import Node

from qoala.runtime.environment import GlobalEnvironment


class HostComponent(Component):
    """NetSquid component representing a Host.

    Subcomponent of a ProcNodeComponent.

    This is a static container for Host-related components and ports. Behavior
    of a Host is modeled in the `Host` class, which is a subclass of `Protocol`.
    """

    def __init__(self, node: Node, global_env: GlobalEnvironment) -> None:
        super().__init__(f"{node.name}_host")

        self._peer_in_ports: Dict[str, str] = {}  # peer name -> port name
        self._peer_out_ports: Dict[str, str] = {}  # peer name -> port name

        for other_node in global_env.get_nodes().values():
            if other_node.name == node.name:
                continue
            port_in_name = f"peer_{other_node.name}_in"
            port_out_name = f"peer_{other_node.name}_out"
            self._peer_in_ports[other_node.name] = port_in_name
            self._peer_out_ports[other_node.name] = port_out_name

        self.add_ports(self._peer_in_ports.values())
        self.add_ports(self._peer_out_ports.values())

        self.add_ports(["qnos_in", "qnos_out"])

    @property
    def qnos_in_port(self) -> Port:
        return self.ports["qnos_in"]

    @property
    def qnos_out_port(self) -> Port:
        return self.ports["qnos_out"]

    def peer_in_port(self, name: str) -> Port:
        port_name = self._peer_in_ports[name]
        return self.ports[port_name]

    def peer_out_port(self, name: str) -> Port:
        port_name = self._peer_out_ports[name]
        return self.ports[port_name]



from __future__ import annotations

import os
from dataclasses import dataclass
from typing import Any, Dict, Generator, List, Optional, Tuple

import netsquid as ns
from netqasm.lang.instr.core import CreateEPRInstruction, RecvEPRInstruction
from netsquid.components import QuantumProcessor
from netsquid.nodes import Node
from netsquid.qubits import ketstates, qubitapi
from netsquid_magic.link_layer import (
    MagicLinkLayerProtocolWithSignaling,
    SingleClickTranslationUnit,
)
from netsquid_magic.magic_distributor import PerfectStateMagicDistributor

from pydynaa import EventExpression
from qoala.lang.iqoala import IqoalaParser, IqoalaProgram
from qoala.runtime.config import GenericQDeviceConfig
from qoala.runtime.environment import GlobalEnvironment, GlobalNodeInfo
from qoala.runtime.program import ProgramInput, ProgramInstance, ProgramResult
from qoala.runtime.schedule import (
    ProgramTaskList,
    Schedule,
    ScheduleEntry,
    ScheduleTime,
    TaskBuilder,
)
from qoala.sim.build import build_generic_qprocessor
from qoala.sim.csocket import ClassicalSocket
from qoala.sim.egp import EgpProtocol
from qoala.sim.hostinterface import HostInterface
from qoala.sim.memory import ProgramMemory, UnitModule
from qoala.sim.process import IqoalaProcess
from qoala.sim.procnode import ProcNode


def create_process(
    pid: int,
    program: IqoalaProgram,
    unit_module: UnitModule,
    host_interface: HostInterface,
    inputs: Optional[Dict[str, Any]] = None,
    tasks: Optional[ProgramTaskList] = None,
) -> IqoalaProcess:
    if inputs is None:
        inputs = {}
    if tasks is None:
        tasks = ProgramTaskList.empty(program)
    prog_input = ProgramInput(values=inputs)
    instance = ProgramInstance(
        pid=pid,
        program=program,
        inputs=prog_input,
        tasks=tasks,
    )
    mem = ProgramMemory(pid=0, unit_module=unit_module)

    process = IqoalaProcess(
        prog_instance=instance,
        prog_memory=mem,
        csockets={
            id: ClassicalSocket(host_interface, name)
            for (id, name) in program.meta.csockets.items()
        },
        epr_sockets=program.meta.epr_sockets,
        subroutines=program.subroutines,
        requests=program.requests,
        result=ProgramResult(values={}),
    )
    return process


def create_qprocessor(name: str, num_qubits: int) -> QuantumProcessor:
    cfg = GenericQDeviceConfig.perfect_config(num_qubits=num_qubits)
    return build_generic_qprocessor(name=f"{name}_processor", cfg=cfg)


def create_global_env(
    num_qubits: int, names: List[str] = ["alice", "bob", "charlie"]
) -> GlobalEnvironment:

    env = GlobalEnvironment()
    for i, name in enumerate(names):
        env.add_node(i, GlobalNodeInfo(name, i))
    return env


def create_procnode(
    name: str,
    env: GlobalEnvironment,
    num_qubits: int,
    asynchronous: bool = False,
) -> ProcNode:
    alice_qprocessor = create_qprocessor(name, num_qubits)

    node_id = env.get_node_id(name)
    procnode = ProcNode(
        name=name,
        global_env=env,
        qprocessor=alice_qprocessor,
        node_id=node_id,
        asynchronous=asynchronous,
    )

    return procnode


def create_egp_protocols(node1: Node, node2: Node) -> Tuple[EgpProtocol, EgpProtocol]:
    link_dist = PerfectStateMagicDistributor(nodes=[node1, node2], state_delay=1000.0)
    link_prot = MagicLinkLayerProtocolWithSignaling(
        nodes=[node1, node2],
        magic_distributor=link_dist,
        translation_unit=SingleClickTranslationUnit(),
    )
    return EgpProtocol(node1, link_prot), EgpProtocol(node2, link_prot)


class BqcProcNode(ProcNode):
    def run_epr_subroutine(
        self, process: IqoalaProcess, subrt_name: str
    ) -> Generator[EventExpression, None, None]:
        subrt = process.subroutines[subrt_name]
        epr_instr_idx = None
        for i, instr in enumerate(subrt.subroutine.instructions):
            if isinstance(instr, CreateEPRInstruction) or isinstance(
                instr, RecvEPRInstruction
            ):
                epr_instr_idx = i
                break

        # Set up arrays
        for i in range(epr_instr_idx):
            yield from self.qnos.processor.assign(process, subrt_name, i)

        request_name = subrt.request_name
        assert request_name is not None
        request = process.requests[request_name].request

        # Handle request
        yield from self.netstack.processor.assign(process, request)

        # Execute wait instruction
        yield from self.qnos.processor.assign(process, subrt_name, epr_instr_idx + 1)

        # Return subroutine results
        self.host.processor.copy_subroutine_results(process, subrt_name)

    def run_subroutine(
        self, process: IqoalaProcess, subrt_name: str
    ) -> Generator[EventExpression, None, None]:
        subrt = process.subroutines[subrt_name]
        num_instrs = len(subrt.subroutine.instructions)

        for i in range(num_instrs):
            yield from self.qnos.processor.assign(process, subrt_name, i)

        # Return subroutine results
        self.host.processor.copy_subroutine_results(process, subrt_name)


def create_server_tasks(server_program: IqoalaProgram) -> ProgramTaskList:
    tasks = []

    cl_dur = 1e3
    cc_dur = 10e6
    ql_dur = 1e3
    qc_dur = 1e6

    tasks.append(TaskBuilder.CL(cl_dur, 0))
    tasks.append(TaskBuilder.CL(cl_dur, 1))
    tasks.append(TaskBuilder.QC(qc_dur, "create_epr_0"))
    tasks.append(TaskBuilder.CL(cl_dur, 2))
    tasks.append(TaskBuilder.QC(qc_dur, "create_epr_1"))
    tasks.append(TaskBuilder.CL(cl_dur, 3))
    tasks.append(TaskBuilder.QL(ql_dur, "local_cphase", 0))
    tasks.append(TaskBuilder.QL(ql_dur, "local_cphase", 1))
    tasks.append(TaskBuilder.QL(ql_dur, "local_cphase", 2))
    tasks.append(TaskBuilder.CC(cc_dur, 4))
    tasks.append(TaskBuilder.CL(cl_dur, 5))
    tasks.append(TaskBuilder.QL(ql_dur, "meas_qubit_1", 0))
    tasks.append(TaskBuilder.QL(ql_dur, "meas_qubit_1", 1))
    tasks.append(TaskBuilder.QL(ql_dur, "meas_qubit_1", 2))
    tasks.append(TaskBuilder.QL(ql_dur, "meas_qubit_1", 3))
    tasks.append(TaskBuilder.QL(ql_dur, "meas_qubit_1", 4))
    tasks.append(TaskBuilder.CC(cc_dur, 6))
    tasks.append(TaskBuilder.CC(cc_dur, 7))
    tasks.append(TaskBuilder.CL(cl_dur, 8))
    tasks.append(TaskBuilder.QL(ql_dur, "meas_qubit_0", 0))
    tasks.append(TaskBuilder.QL(ql_dur, "meas_qubit_0", 1))
    tasks.append(TaskBuilder.QL(ql_dur, "meas_qubit_0", 2))
    tasks.append(TaskBuilder.QL(ql_dur, "meas_qubit_0", 3))
    tasks.append(TaskBuilder.QL(ql_dur, "meas_qubit_0", 4))
    tasks.append(TaskBuilder.CL(cl_dur, 9))
    tasks.append(TaskBuilder.CL(cl_dur, 10))

    return ProgramTaskList(server_program, {i: task for i, task in enumerate(tasks)})


def create_client_tasks(client_program: IqoalaProgram) -> ProgramTaskList:
    tasks = []

    cl_dur = 1e3
    cc_dur = 10e6
    ql_dur = 1e3
    qc_dur = 1e6

    tasks.append(TaskBuilder.CL(cl_dur, 0))
    tasks.append(TaskBuilder.CL(cl_dur, 1))
    tasks.append(TaskBuilder.QC(qc_dur, "create_epr_0"))
    tasks.append(TaskBuilder.CL(cl_dur, 2))
    tasks.append(TaskBuilder.QL(ql_dur, "post_epr_0", 0))
    tasks.append(TaskBuilder.QL(ql_dur, "post_epr_0", 1))
    tasks.append(TaskBuilder.QL(ql_dur, "post_epr_0", 2))
    tasks.append(TaskBuilder.QL(ql_dur, "post_epr_0", 3))
    tasks.append(TaskBuilder.QL(ql_dur, "post_epr_0", 4))

    tasks.append(TaskBuilder.CL(cl_dur, 3))
    tasks.append(TaskBuilder.QC(qc_dur, "create_epr_1"))
    tasks.append(TaskBuilder.CL(cl_dur, 4))
    tasks.append(TaskBuilder.QL(ql_dur, "post_epr_1", 0))
    tasks.append(TaskBuilder.QL(ql_dur, "post_epr_1", 1))
    tasks.append(TaskBuilder.QL(ql_dur, "post_epr_1", 2))
    tasks.append(TaskBuilder.QL(ql_dur, "post_epr_1", 3))
    tasks.append(TaskBuilder.QL(ql_dur, "post_epr_1", 4))

    tasks.append(TaskBuilder.CL(cl_dur, 5))
    tasks.append(TaskBuilder.CL(cl_dur, 6))
    tasks.append(TaskBuilder.CL(cl_dur, 7))
    tasks.append(TaskBuilder.CL(cl_dur, 8))
    tasks.append(TaskBuilder.CC(cc_dur, 9))
    tasks.append(TaskBuilder.CC(cc_dur, 10))
    tasks.append(TaskBuilder.CL(cl_dur, 11))
    tasks.append(TaskBuilder.CL(cl_dur, 12))
    tasks.append(TaskBuilder.CL(cl_dur, 13))
    tasks.append(TaskBuilder.CL(cl_dur, 14))
    tasks.append(TaskBuilder.CL(cl_dur, 15))
    tasks.append(TaskBuilder.CC(cc_dur, 16))
    tasks.append(TaskBuilder.CL(cl_dur, 17))
    tasks.append(TaskBuilder.CL(cl_dur, 18))

    return ProgramTaskList(client_program, {i: task for i, task in enumerate(tasks)})


def create_server_schedule(num_tasks: int) -> Schedule:
    entries = [
        (ScheduleTime(int(i * 1e8)), ScheduleEntry(pid=0, task_index=i))
        for i in range(num_tasks)
    ]
    return Schedule(entries)


def create_client_schedule(num_tasks: int) -> Schedule:
    entries = [
        (ScheduleTime(None), ScheduleEntry(pid=0, task_index=i))
        for i in range(num_tasks)
    ]
    return Schedule(entries)


class ServerProcNode(BqcProcNode):
    def run(self) -> Generator[EventExpression, None, None]:
        self.finished = False

        self.finished = True


class ClientProcNode(BqcProcNode):
    def run(self) -> Generator[EventExpression, None, None]:
        self.finished = False

        self.finished = True


@dataclass
class BqcResult:
    client_process: IqoalaProcess
    server_process: IqoalaProcess
    client_procnode: BqcProcNode
    server_procnode: BqcProcNode


def run_bqc(
    alpha,
    beta,
    theta1,
    theta2,
):
    num_qubits = 3
    unit_module = UnitModule.default_generic(num_qubits)
    global_env = create_global_env(num_qubits, names=["client", "server"])
    server_id = global_env.get_node_id("server")
    client_id = global_env.get_node_id("client")

    path = os.path.join(os.path.dirname(__file__), "test_bqc_server.iqoala")
    with open(path) as file:
        server_text = file.read()
    server_program = IqoalaParser(server_text).parse()
    server_tasks = create_server_tasks(server_program)

    server_procnode = create_procnode("server", global_env, num_qubits)
    server_process = create_process(
        pid=0,
        program=server_program,
        unit_module=unit_module,
        host_interface=server_procnode.host._interface,
        inputs={"client_id": client_id},
        tasks=server_tasks,
    )
    server_procnode.add_process(server_process)
    server_procnode.scheduler.initialize_process(server_process)

    server_schedule = create_server_schedule(len(server_tasks.tasks))
    server_procnode.install_schedule(server_schedule)

    path = os.path.join(os.path.dirname(__file__), "test_bqc_client.iqoala")
    with open(path) as file:
        client_text = file.read()
    client_program = IqoalaParser(client_text).parse()
    client_tasks = create_client_tasks(client_program)

    client_procnode = create_procnode("client", global_env, num_qubits)
    client_process = create_process(
        pid=0,
        program=client_program,
        unit_module=unit_module,
        host_interface=client_procnode.host._interface,
        inputs={
            "server_id": server_id,
            "alpha": alpha,
            "beta": beta,
            "theta1": theta1,
            "theta2": theta2,
        },
        tasks=client_tasks,
    )
    client_procnode.add_process(client_process)
    client_procnode.scheduler.initialize_process(client_process)

    client_schedule = create_client_schedule(len(client_tasks.tasks))
    client_procnode.install_schedule(client_schedule)

    client_egp, server_egp = create_egp_protocols(
        client_procnode.node, server_procnode.node
    )
    client_procnode.egpmgr.add_egp(server_id, client_egp)
    server_procnode.egpmgr.add_egp(client_id, server_egp)

    client_procnode.connect_to(server_procnode)

    server_procnode.start()
    client_procnode.start()
    ns.sim_run()

    return BqcResult(
        client_process=client_process,
        server_process=server_process,
        client_procnode=client_procnode,
        server_procnode=server_procnode,
    )


def expected_rsp_qubit(theta: int, p: int, dummy: bool):
    expected = qubitapi.create_qubits(1)[0]

    if dummy:
        if p == 0:
            qubitapi.assign_qstate(expected, ketstates.s0)
        elif p == 1:
            qubitapi.assign_qstate(expected, ketstates.s1)
    else:
        if (theta, p) == (0, 0):
            qubitapi.assign_qstate(expected, ketstates.h0)
        elif (theta, p) == (0, 1):
            qubitapi.assign_qstate(expected, ketstates.h1)
        if (theta, p) == (8, 0):
            qubitapi.assign_qstate(expected, ketstates.y0)
        elif (theta, p) == (8, 1):
            qubitapi.assign_qstate(expected, ketstates.y1)
        if (theta, p) == (16, 0):
            qubitapi.assign_qstate(expected, ketstates.h1)
        elif (theta, p) == (16, 1):
            qubitapi.assign_qstate(expected, ketstates.h0)
        if (theta, p) == (-8, 0):
            qubitapi.assign_qstate(expected, ketstates.y1)
        elif (theta, p) == (-8, 1):
            qubitapi.assign_qstate(expected, ketstates.y0)

    return expected


def expected_rsp_state(theta: int, p: int, dummy: bool):
    expected = qubitapi.create_qubits(1)[0]

    if dummy:
        if p == 0:
            return ketstates.s0
        elif p == 1:
            return ketstates.s1
    else:
        if (theta, p) == (0, 0):
            return ketstates.h0
        elif (theta, p) == (0, 1):
            return ketstates.h1
        if (theta, p) == (8, 0):
            return ketstates.y0
        elif (theta, p) == (8, 1):
            return ketstates.y1
        if (theta, p) == (16, 0):
            return ketstates.h1
        elif (theta, p) == (16, 1):
            return ketstates.h0
        if (theta, p) == (-8, 0):
            return ketstates.y1
        elif (theta, p) == (-8, 1):
            return ketstates.y0

    return expected.qstate


def test_bqc():

    # Effective computation: measure in Z the following state:
    # H Rz(beta) H Rz(alpha) |+>
    # m2 should be this outcome

    # angles are in multiples of pi/16

    def check(alpha, beta, theta1, theta2, expected):
        ns.sim_reset()
        results = [
            run_bqc(
                alpha=alpha,
                beta=beta,
                theta1=theta1,
                theta2=theta2,
            )
            for _ in range(1)
        ]
        m2s = [result.server_process.result.values["m2"] for result in results]
        assert all(m2 == expected for m2 in m2s)

    check(alpha=8, beta=8, theta1=0, theta2=0, expected=0)
    check(alpha=8, beta=24, theta1=0, theta2=0, expected=1)
    check(alpha=8, beta=8, theta1=13, theta2=27, expected=0)
    check(alpha=8, beta=24, theta1=2, theta2=22, expected=1)


if __name__ == "__main__":
    test_bqc()



from __future__ import annotations

import logging
from copy import deepcopy
from typing import Dict, Generator, List, Optional, Tuple, Type

import netsquid as ns
from netqasm.lang.instr.core import CreateEPRInstruction, RecvEPRInstruction
from netqasm.lang.operand import Template
from netsquid.protocols import Protocol

from pydynaa import EventExpression
from qoala.runtime.environment import LocalEnvironment
from qoala.runtime.program import (
    BatchInfo,
    BatchResult,
    ProgramBatch,
    ProgramInstance,
    ProgramResult,
)
from qoala.runtime.schedule import (
    CombinedProgramTask,
    HostTask,
    NetstackTask,
    ProcessorType,
    ProgramTask,
    ProgramTaskList,
    QnosTask,
    Schedule,
    ScheduleEntry,
    SchedulerInput,
    SchedulerOutput,
    ScheduleSolver,
    ScheduleTime,
    SingleProgramTask,
)
from qoala.sim.csocket import ClassicalSocket
from qoala.sim.eprsocket import EprSocket
from qoala.sim.events import EVENT_WAIT
from qoala.sim.host import Host
from qoala.sim.logging import LogManager
from qoala.sim.memmgr import MemoryManager
from qoala.sim.memory import ProgramMemory, UnitModule
from qoala.sim.netstack import Netstack
from qoala.sim.process import IqoalaProcess
from qoala.sim.qnos import Qnos


class Scheduler(Protocol):
    def __init__(
        self,
        node_name: str,
        host: Host,
        qnos: Qnos,
        netstack: Netstack,
        memmgr: MemoryManager,
        local_env: LocalEnvironment,
    ) -> None:
        super().__init__(name=f"{node_name}_scheduler")

        self._node_name = node_name

        self._logger: logging.Logger = LogManager.get_stack_logger(  # type: ignore
            f"{self.__class__.__name__}({node_name})"
        )

        self._host = host
        self._qnos = qnos
        self._netstack = netstack
        self._memmgr = memmgr
        self._local_env = local_env

        self._prog_instance_counter: int = 0
        self._batch_counter: int = 0
        self._batches: Dict[int, ProgramBatch] = {}  # batch ID -> batch
        self._prog_results: Dict[int, ProgramResult] = {}  # program ID -> result
        self._batch_results: Dict[int, BatchResult] = {}  # batch ID -> result

        self._schedule: Optional[Schedule] = None

    @property
    def host(self) -> Host:
        return self._host

    @property
    def qnos(self) -> Qnos:
        return self._qnos

    @property
    def netstack(self) -> Netstack:
        return self._netstack

    @property
    def memmgr(self) -> MemoryManager:
        return self._memmgr

    def submit_batch(self, batch_info: BatchInfo) -> None:
        prog_instances: List[ProgramInstance] = []

        for i in range(batch_info.num_iterations):
            instance = ProgramInstance(
                pid=self._prog_instance_counter,
                program=batch_info.program,
                inputs=batch_info.inputs[i],
                tasks=batch_info.tasks,
            )
            self._prog_instance_counter += 1
            prog_instances.append(instance)

        batch = ProgramBatch(
            batch_id=self._batch_counter, info=batch_info, instances=prog_instances
        )
        self._batches[batch.batch_id] = batch
        self._batch_counter += 1

    def get_batches(self) -> Dict[int, ProgramBatch]:
        return self._batches

    def get_tasks_to_schedule(self) -> SchedulerInput:
        num_programs = len(self._batches)
        deadlines = [b.info.deadline for b in self._batches.values()]
        num_executions = [b.info.num_iterations for b in self._batches.values()]
        tasks: Dict[int, ProgramTaskList]  # batch ID -> task list
        tasks = {i: b.info.tasks for i, b in self._batches.items()}
        num_instructions = [len(task.tasks) for _, task in tasks.items()]
        instr_durations = [
            [t.duration for t in task.tasks.values()] for _, task in tasks.items()
        ]
        instr_types = [
            [t.instr_type for t in task.tasks.values()] for _, task in tasks.items()
        ]

        global_schedule = self._local_env.get_global_env().get_global_schedule()
        timeslot_len = self._local_env.get_global_env().get_timeslot_len()

        return SchedulerInput(
            global_schedule=global_schedule,
            timeslot_len=timeslot_len,
            num_programs=num_programs,
            deadlines=deadlines,
            num_executions=num_executions,
            num_instructions=num_instructions,
            instr_durations=instr_durations,
            instr_types=instr_types,
        )

    def solver_output_to_schedule(self, output: SchedulerOutput) -> Schedule:
        schedule_entries: List[Tuple[ScheduleTime, ScheduleEntry]] = []
        for entry in output.entries:
            batch_id = entry.app_index
            batch = self._batches[batch_id]
            instance_idx = entry.ex_index
            prog_instance = batch.instances[instance_idx]
            pid = prog_instance.pid
            task_index = entry.instr_index
            time = entry.start_time  # note: may be None!
            schedule_entries.append(
                (ScheduleTime(time), ScheduleEntry(pid, task_index))
            )
        return Schedule(schedule_entries)

    def solve_and_install_schedule(self, solver: Type[ScheduleSolver]) -> None:
        solver_input = self.get_tasks_to_schedule()
        solver_output = solver.solve(solver_input)
        schedule = self.solver_output_to_schedule(solver_output)
        self.install_schedule(schedule)

    def create_processes_for_batches(self) -> None:
        for batch in self._batches.values():
            for prog_instance in batch.instances:
                prog_memory = ProgramMemory(
                    prog_instance.pid,
                    unit_module=UnitModule.default_generic(batch.info.num_qubits),
                )
                meta = prog_instance.program.meta

                csockets: Dict[int, ClassicalSocket] = {}
                for i, remote_name in meta.csockets.items():
                    # TODO: check for already existing epr sockets
                    csockets[i] = self.host.create_csocket(remote_name)

                epr_sockets: Dict[int, EprSocket] = {}
                for i, remote_name in meta.epr_sockets.items():
                    global_env = self._local_env.get_global_env()
                    remote_id = global_env.get_node_id(remote_name)
                    # TODO: check for already existing epr sockets
                    # TODO: fidelity
                    epr_sockets[i] = EprSocket(i, remote_id, 1.0)

                result = ProgramResult(values={})

                # Important: create a deep copy of the subroutines for each process,
                # since each process should be able to instantiate their subroutines
                # without affecting the subroutines of other processes.
                subroutines = {
                    name: deepcopy(subrt)
                    for name, subrt in prog_instance.program.subroutines.items()
                }

                # Same holds for requests.
                requests = {
                    name: deepcopy(req)
                    for name, req in prog_instance.program.requests.items()
                }

                process = IqoalaProcess(
                    prog_instance=prog_instance,
                    prog_memory=prog_memory,
                    csockets=csockets,
                    epr_sockets=epr_sockets,
                    subroutines=subroutines,
                    requests=requests,
                    result=result,
                )

                self.memmgr.add_process(process)
                self.initialize_process(process)

    def collect_batch_results(self) -> None:
        for batch_id, batch in self._batches.items():
            results: List[ProgramResult] = []
            for prog_instance in batch.instances:
                process = self.memmgr.get_process(prog_instance.pid)
                results.append(process.result)
            self._batch_results[batch_id] = BatchResult(batch_id, results)

    def get_batch_results(self) -> Dict[int, BatchResult]:
        return self._batch_results

    def execute_host_task(
        self, process: IqoalaProcess, task: HostTask
    ) -> Generator[EventExpression, None, None]:
        yield from self.host.processor.assign(process, task.instr_index)

    def execute_qnos_task(
        self, process: IqoalaProcess, task: QnosTask
    ) -> Generator[EventExpression, None, None]:
        yield from self.qnos.processor.assign(
            process, task.subrt_name, task.instr_index
        )
        # TODO: improve this
        subrt = process.subroutines[task.subrt_name]
        if task.instr_index == (len(subrt.subroutine.instructions) - 1):
            # subroutine finished -> return results to host
            self.host.processor.copy_subroutine_results(process, task.subrt_name)

    def run_epr_subroutine(
        self, process: IqoalaProcess, subrt_name: str
    ) -> Generator[EventExpression, None, None]:
        subrt = process.subroutines[subrt_name]
        epr_instr_idx: Optional[int] = None
        for i, instr in enumerate(subrt.subroutine.instructions):
            if isinstance(instr, CreateEPRInstruction) or isinstance(
                instr, RecvEPRInstruction
            ):
                epr_instr_idx = i
                break

        assert epr_instr_idx is not None

        # Set up arrays
        for i in range(epr_instr_idx):
            yield from self.qnos.processor.assign(process, subrt_name, i)

        request_name = subrt.request_name
        assert request_name is not None
        request = process.requests[request_name].request

        # Handle request
        yield from self.netstack.processor.assign(process, request)

        # Execute wait instruction
        yield from self.qnos.processor.assign(process, subrt_name, epr_instr_idx + 1)

        # Return subroutine results
        self.host.processor.copy_subroutine_results(process, subrt_name)

    def execute_netstack_task(
        self, process: IqoalaProcess, task: NetstackTask
    ) -> Generator[EventExpression, None, None]:
        yield from self.run_epr_subroutine(process, task.subrt_name)

    def execute_single_task(
        self, process: IqoalaProcess, task: SingleProgramTask
    ) -> Generator[EventExpression, None, None]:
        if task.processor_type == ProcessorType.HOST:
            yield from self.execute_host_task(process, task.as_host_task())
        elif task.processor_type == ProcessorType.QNOS:
            yield from self.execute_qnos_task(process, task.as_qnos_task())
        elif task.processor_type == ProcessorType.NETSTACK:
            yield from self.execute_netstack_task(process, task.as_netstack_task())
        else:
            raise RuntimeError

    def execute_task(
        self, process: IqoalaProcess, task: ProgramTask
    ) -> Generator[EventExpression, None, None]:
        if isinstance(task, SingleProgramTask):
            yield from self.execute_single_task(process, task)
        elif isinstance(task, CombinedProgramTask):
            for task in task.tasks:
                yield from self.execute_single_task(process, task)
        else:
            raise RuntimeError

    def initialize_process(self, process: IqoalaProcess) -> None:
        # Write program inputs to host memory.
        self.host.processor.initialize(process)

        inputs = process.prog_instance.inputs
        for req in process.requests.values():
            # TODO: support for other request parameters being templates?
            remote_id = req.request.remote_id
            if isinstance(remote_id, Template):
                req.request.remote_id = inputs.values[remote_id.name]

    def install_schedule(self, schedule: Schedule) -> None:
        self._schedule = schedule

    def wait(self, delta_time: int) -> Generator[EventExpression, None, None]:
        self._schedule_after(delta_time, EVENT_WAIT)
        event_expr = EventExpression(source=self, event_type=EVENT_WAIT)
        yield event_expr

    def run(self) -> Generator[EventExpression, None, None]:
        if self._schedule is None:
            return

        for schedule_time, entry in self._schedule.entries:
            process = self.memmgr.get_process(entry.pid)
            task_list = process.prog_instance.tasks
            task = task_list.tasks[entry.task_index]

            if schedule_time.time is None:  # no time constraint
                # print(f"{self.name} executing task {task}")
                yield from self.execute_task(process, task)
            else:
                ns_time = ns.sim_time()
                delta = schedule_time.time - ns.sim_time()
                self._logger.debug(
                    f"{self.name} next scheduled time = {schedule_time.time}, delta = {delta}"
                )
                yield from self.wait(delta)
                ns_time = ns.sim_time()
                self._logger.debug(
                    f"{self.name} ns_time: {ns_time}, executing task {task}"
                )
                yield from self.execute_task(process, task)
                ns_time = ns.sim_time()
                self._logger.debug(
                    f"{self.name} ns_time: {ns_time}, finished task {task}"
                )

        self._logger.debug(f"{self.name} finished with schedule\n\n")
        self.collect_batch_results()



from __future__ import annotations

from typing import Dict, List

from netsquid.components import QuantumProcessor
from netsquid.nodes.network import Network
from netsquid_magic.link_layer import MagicLinkLayerProtocol

from qoala.sim.procnode import ProcNode


class ProcNodeNetwork(Network):
    """A network of `ProcNode`s connected by links, which are
    `MagicLinkLayerProtocol`s."""

    def __init__(
        self, nodes: Dict[str, ProcNode], links: List[MagicLinkLayerProtocol]
    ) -> None:
        """ProcNodeNetwork constructor.

        :param nodes: dictionary of node name to `ProcNode` object representing
        that node
        :param links: list of link layer protocol objects. Each object internally
        contains the IDs of the two nodes that this link connects
        """
        self._nodes = nodes
        self._links = links

    @property
    def nodes(self) -> Dict[str, ProcNode]:
        return self._nodes

    @property
    def links(self) -> List[MagicLinkLayerProtocol]:
        return self._links

    @property
    def qdevices(self) -> Dict[str, QuantumProcessor]:
        return {name: node.qdevice for name, node in self._nodes.items()}

    def start_all_nodes(self) -> None:
        for node in self.nodes.values():
            node.start()



from __future__ import annotations

from dataclasses import dataclass
from typing import Dict, List, Optional, Tuple

from qoala.runtime.config import LinkConfig
from qoala.runtime.program import ProgramInstance


@dataclass
class GlobalNodeInfo:
    """Node information available at runtime."""

    name: str
    id: int


@dataclass
class GlobalLinkInfo:
    node_name1: str
    node_name2: str

    fidelity: float

    @classmethod
    def from_config(
        cls, node_name1: str, node_name2: str, config: LinkConfig
    ) -> GlobalLinkInfo:
        if config.typ == "perfect":
            return GlobalLinkInfo(
                node_name1=node_name1, node_name2=node_name2, fidelity=1.0
            )
        elif config.typ == "depolarise":
            return GlobalLinkInfo(
                node_name1=node_name1,
                node_name2=node_name2,
                fidelity=config.cfg.fidelity,  # type: ignore
            )
        else:
            raise NotImplementedError


class GlobalEnvironment:
    def __init__(self) -> None:
        # node ID -> node info
        self._nodes: Dict[int, GlobalNodeInfo] = {}

        # (node A ID, node B ID) -> link info
        # for a pair (a, b) there exists no separate (b, a) info (it is the same)
        self._links: Dict[Tuple[int, int], GlobalLinkInfo] = {}

        self._global_schedule: Optional[List[int]] = None
        self._timeslot_len: Optional[int] = None

    def get_nodes(self) -> Dict[int, GlobalNodeInfo]:
        return self._nodes

    def get_node_id(self, name: str) -> int:
        for id, node in self._nodes.items():
            if node.name == name:
                return id
        raise ValueError

    def set_nodes(self, nodes: Dict[int, GlobalNodeInfo]) -> None:
        self._nodes = nodes

    def add_node(self, id: int, node: GlobalNodeInfo) -> None:
        self._nodes[id] = node

    def get_links(self) -> Dict[Tuple[int, int], GlobalLinkInfo]:
        return self._links

    def set_links(self, links: Dict[Tuple[int, int], GlobalLinkInfo]) -> None:
        self._links = links

    def add_link(self, id1: int, id2: int, link: GlobalLinkInfo) -> None:
        self._links[(id1, id2)] = link

    def set_global_schedule(self, schedule: List[int]) -> None:
        self._global_schedule = schedule

    def get_global_schedule(self) -> List[int]:
        assert self._global_schedule is not None
        return self._global_schedule

    def set_timeslot_len(self, len: int) -> None:
        self._timeslot_len = len

    def get_timeslot_len(self) -> int:
        assert self._timeslot_len is not None
        return self._timeslot_len


class LocalEnvironment:
    def __init__(
        self,
        global_env: GlobalEnvironment,
        node_id: int,
    ) -> None:
        self._global_env: GlobalEnvironment = global_env

        # node ID of self
        self._node_id: int = node_id

        self._programs: List[ProgramInstance] = []
        self._csockets: List[str] = []
        self._epr_sockets: List[str] = []

    def get_global_env(self) -> GlobalEnvironment:
        return self._global_env

    def get_node_id(self) -> int:
        return self._node_id

    def register_program(self, program: ProgramInstance) -> None:
        self._programs.append(program)

    def open_epr_socket(self) -> None:
        pass

    def get_all_node_names(self) -> List[str]:
        return [info.name for info in self.get_global_env().get_nodes().values()]

    def get_all_other_node_names(self) -> List[str]:
        return [
            info.name
            for info in self.get_global_env().get_nodes().values()
            if info.id != self._node_id
        ]


class ProgramEnvironment:
    """Environment interface given to a program"""

    pass



from dataclasses import dataclass
from typing import Dict

from qoala.lang.iqoala import IqoalaRequest, IqoalaSubroutine
from qoala.runtime.program import ProgramInstance, ProgramResult
from qoala.sim.csocket import ClassicalSocket
from qoala.sim.eprsocket import EprSocket
from qoala.sim.memory import HostMemory, ProgramMemory, SharedMemory


@dataclass
class IqoalaProcess:
    prog_instance: ProgramInstance

    # Mutable
    prog_memory: ProgramMemory
    result: ProgramResult

    # Immutable
    csockets: Dict[int, ClassicalSocket]
    epr_sockets: Dict[int, EprSocket]
    subroutines: Dict[str, IqoalaSubroutine]
    requests: Dict[str, IqoalaRequest]

    @property
    def pid(self) -> int:
        return self.prog_instance.pid  # type: ignore

    @property
    def host_mem(self) -> HostMemory:
        return self.prog_memory.host_mem

    @property
    def shared_mem(self) -> SharedMemory:
        return self.prog_memory.shared_mem



from dataclasses import dataclass
from enum import Enum, auto
from typing import List, Optional, Union


class EprCreateType(Enum):
    CREATE_KEEP = 0
    MEASURE_DIRECTLY = auto()
    REMOTE_STATE_PREP = auto()


class EprCreateRole(Enum):
    CREATE = 0
    RECEIVE = auto()


@dataclass
class NetstackCreateRequest:
    # Request parameters.
    remote_id: int
    epr_socket_id: int
    typ: EprCreateType
    num_pairs: int
    fidelity: float
    virt_qubit_ids: List[int]

    # Info for writing results.
    result_array_addr: int


@dataclass
class NetstackReceiveRequest:
    # Request parameters.
    remote_id: int
    epr_socket_id: int
    typ: Optional[EprCreateType]  # not knowable from recv_epr instruction! TODO
    num_pairs: Optional[int]  # not knowable from recv_epr instruction! TODO
    fidelity: float
    virt_qubit_ids: List[int]

    # Info for writing results.
    result_array_addr: int


@dataclass
class NetstackBreakpointCreateRequest:
    pid: int


@dataclass
class NetstackBreakpointReceiveRequest:
    pid: int


T_NetstackRequest = Union[
    NetstackCreateRequest,
    NetstackReceiveRequest,
    NetstackBreakpointCreateRequest,
    NetstackBreakpointReceiveRequest,
]
